from typing_extensions import override

from torch import Tensor, nn

from conduit.types import Loss

__all__ = ["OnlineReweightingLoss"]


class OnlineReweightingLoss(nn.Module):
    """Wrapper that computes a loss balanced by intersectional group size."""

    def __init__(self, loss_fn: Loss) -> None:
        super().__init__()
        # the base loss function needs to produce instance-wise losses for the
        # reweighting (determined by subgroup cardinality) to be applied
        loss_fn.reduction = "none"
        self.loss_fn = loss_fn

    @override
    def forward(self, logits: Tensor, targets: Tensor, subgroup_inf: Tensor) -> Tensor:
        unweighted_loss = self.loss_fn(logits, targets)
        for _y in targets.unique():
            for _s in subgroup_inf.unique():
                # compute the cardinality of each subgroup and use this to weight the sample-losses
                mask = (targets == _y) & (subgroup_inf == _s)
                unweighted_loss[mask] /= mask.sum()
        return unweighted_loss.sum()



from typing import List, Optional, Union

import numpy as np
import numpy.typing as npt
from ranzen import some
import torch
from torch.functional import Tensor

from conduit.data.datasets.base import CdtDataset, I, S, Y
from conduit.data.structures import TargetData
from conduit.transforms.tabular import TabularNormalize

__all__ = ["CdtTabularDataset"]


class CdtTabularDataset(CdtDataset[I, Tensor, Y, S]):
    """A dataset for tabular data.

    :param x: The input features.
    :param y: The target data.
    :param s: The sensitive attribute data.
    :param non_ohe_indexes: The indexes of the non-one-hot-encoded features. If None, all features
        are assumed to be one-hot-encoded. Transformations will only be applied to the
        non-one-hot-encoded features.
    :param feature_groups: List of one-hot-encoded feature groups. Each group is encoded as a slice.
    """

    x: Tensor

    def __init__(
        self,
        *,
        x: Union[Tensor, npt.NDArray[np.floating], npt.NDArray[np.integer]],
        y: Optional[TargetData] = None,
        s: Optional[TargetData] = None,
        non_ohe_indexes: Optional[List[int]] = None,
        feature_groups: Optional[List[slice]] = None,
    ) -> None:
        if isinstance(x, np.ndarray):
            x = torch.as_tensor(x, dtype=torch.float32)
        super().__init__(x=x, y=y, s=s)

        self.non_ohe_indexes = non_ohe_indexes
        self.feature_groups = feature_groups

    @property
    def x_non_ohe(self) -> Optional[Tensor]:
        if self.non_ohe_indexes is None:
            return None
        return self.x[:, self.non_ohe_indexes]

    def fit_transform_(self, transform: TabularNormalize) -> None:
        """Fit a transformation to the non-one-hot-encoded features and transform in-place."""
        if some(x_non_ohe := self.x_non_ohe):
            self.x[:, self.non_ohe_indexes] = transform.fit_transform(x_non_ohe)

    def transform_(self, transform: TabularNormalize) -> None:
        """Transform the non-one-hot-encoded features in-place."""
        if some(x_non_ohe := self.x_non_ohe):
            self.x[:, self.non_ohe_indexes] = transform.transform(x_non_ohe)



"""Camelyon17 data-module."""

from dataclasses import dataclass
from typing import Any
from typing_extensions import override

import albumentations as A  # type: ignore

from conduit.data.datamodules.vision.base import CdtVisionDataModule
from conduit.data.datasets.vision.camelyon17 import (
    Camelyon17,
    Camelyon17Attr,
    Camelyon17Split,
    Camelyon17SplitScheme,
    SampleType,
)
from conduit.data.structures import TrainValTestSplit

__all__ = ["Camelyon17DataModule"]


@dataclass(kw_only=True)
class Camelyon17DataModule(CdtVisionDataModule[SampleType]):
    """Data-module for the Camelyon17 dataset."""

    image_size: int = 96
    superclass: Camelyon17Attr = Camelyon17Attr.TUMOR
    subclass: Camelyon17Attr = Camelyon17Attr.CENTER
    use_predefined_splits: bool = False
    split_scheme: Camelyon17SplitScheme = Camelyon17SplitScheme.OFFICIAL

    @override
    def prepare_data(self, *args: Any, **kwargs: Any) -> None:
        Camelyon17(root=self.root, download=True)

    @property
    @override
    def _default_train_transforms(self) -> A.Compose:
        base_transforms = A.Compose(
            [
                A.Resize(self.image_size, self.image_size),
                A.CenterCrop(self.image_size, self.image_size),
                A.HorizontalFlip(),
                A.RandomCrop(self.image_size, self.image_size),
            ]
        )
        normalization = super()._default_train_transforms

        return A.Compose([base_transforms, normalization])

    @property
    @override
    def _default_test_transforms(self) -> A.Compose:
        base_transforms = A.Compose(
            [
                A.Resize(self.image_size, self.image_size),
                A.CenterCrop(self.image_size, self.image_size),
            ]
        )
        normalization = super()._default_train_transforms

        return A.Compose([base_transforms, normalization])

    @override
    def _get_image_splits(self) -> TrainValTestSplit[Camelyon17]:
        # Split the data according to the pre-defined split indices
        if self.use_predefined_splits:
            train_data, val_data, test_data = (
                Camelyon17(root=self.root, split=split, split_scheme=self.split_scheme)
                for split in Camelyon17Split
            )
        # Split the data randomly according to test- and val-prop
        else:
            all_data = Camelyon17(root=self.root, transform=None, split_scheme=self.split_scheme)
            val_data, test_data, train_data = all_data.random_split(
                props=(self.val_prop, self.test_prop), seed=self.seed
            )
        return TrainValTestSplit(train=train_data, val=val_data, test=test_data)



"""Dataset wrappers."""

from dataclasses import is_dataclass, replace
from typing import Any, Optional, Tuple, Union
from typing_extensions import override

from torch import Tensor

from conduit.data.datasets.utils import AudioTform, apply_audio_transform, compute_instance_weights
from conduit.data.structures import (
    BinarySample,
    BinarySampleIW,
    Dataset,
    DatasetWrapper,
    SampleBase,
    SubgroupSample,
    TernarySample,
    TernarySampleIW,
    _BinarySampleMixin,
    shallow_asdict,
)
from conduit.transforms.tabular import TabularTransform

__all__ = [
    "AudioTransformer",
    "InstanceWeightedDataset",
    "TabularTransformer",
]


class AudioTransformer(DatasetWrapper[Any]):
    """
    Wrapper class for applying image transformations.

    Useful when wanting to have different transformations for different subsets of the data
    that share the same underlying dataset.
    """

    def __init__(self, dataset: Dataset, *, transform: Optional[AudioTform]) -> None:
        self.dataset = dataset
        self._transform: Optional[AudioTform] = None
        self.transform = transform

    @override
    def __getitem__(self, index: int) -> Any:
        sample = self.dataset[index]
        if self.transform is not None:
            if isinstance(sample, SampleBase):
                if not isinstance(sample.x, Tensor):
                    raise TypeError(
                        f"Audio transform cannot be applied to input of type '{type(sample.x)}'"
                        "(must be a PyTorch Tensor)."
                    )
                waveform = apply_audio_transform(waveform=sample.x, transform=self.transform)
                sample = replace(sample, x=waveform)
            elif isinstance(sample, Tensor):
                sample = apply_audio_transform(waveform=sample, transform=self.transform)
            else:
                waveform = apply_audio_transform(waveform=sample[0], transform=self.transform)
                data_type = type(sample)
                sample = data_type(waveform, *sample[1:])
        return sample


class TabularTransformer(DatasetWrapper[Any]):
    """
    Wrapper class for applying transformations to tabular data.

    Useful when wanting to have different transformations for different subsets of the data
    that share the same underlying dataset.
    """

    def __init__(
        self,
        dataset: Dataset,
        *,
        transform: Optional[TabularTransform],
        target_transform: Optional[TabularTransform],
    ) -> None:
        self.dataset = dataset
        self.transform = transform
        self.target_transform = target_transform

    @override
    def __getitem__(self, index: int) -> Any:
        sample = self.dataset[index]
        if self.transform is not None:
            if isinstance(sample, SampleBase):
                if not isinstance(sample.x, Tensor):
                    raise TypeError(
                        f"Tabular transform cannot be applied to input of type '{type(sample.x)}'"
                        "(must be a PyTorch Tensor)."
                    )
                new_values = {"x": self.transform(sample.x)}
                if (self.target_transform is not None) and (isinstance(sample, _BinarySampleMixin)):
                    new_values["y"] = self.target_transform(sample.y)
                sample = replace(sample, **new_values)
            elif isinstance(sample, Tensor):
                sample = self.transform(sample)
            else:
                input_ = self.transform(sample[0])
                data_type = type(sample)
                if (self.target_transform is not None) and len(sample) > 1:
                    target = self.target_transform(sample[1])
                    sample = data_type(input_, target, *sample[2:])
                else:
                    sample = data_type(input_, *sample[1:])
        elif self.target_transform is not None:
            if isinstance(sample, _BinarySampleMixin):
                sample = replace(sample, y=self.target_transform(sample.y))
            elif len(sample) > 1:
                data_type = type(sample)
                target = self.target_transform(sample[1])
                sample = data_type(sample[0], target, *sample[2:])
        return sample


class InstanceWeightedDataset(DatasetWrapper[Any]):
    """Wrapper endowing datasets with instance-weights."""

    def __init__(self, dataset: Dataset) -> None:
        self.dataset = dataset
        self.iw = compute_instance_weights(dataset)

    @override
    def __getitem__(self, index: int) -> Union[SampleBase, Tuple[Any, ...]]:
        sample = self.dataset[index]
        iw = self.iw[index]
        if isinstance(sample, (BinarySample, SubgroupSample, TernarySample)):
            return sample.add_field(iw=iw)
        elif isinstance(sample, tuple):
            if len(sample) == 2:
                x, y = sample
                return BinarySampleIW(x=x, y=y, iw=iw)
            else:
                x, y, s = sample
                return TernarySampleIW(x=x, y=y, s=s, iw=iw)
        elif is_dataclass(sample):
            tuple_class = BinarySampleIW if len(sample) == 2 else TernarySampleIW
            attr_dict = shallow_asdict(sample)
            attr_dict["iw"] = iw  # Covers the corner-case of 'sample' already being an IW sample
            return tuple_class(**attr_dict)
        else:
            raise TypeError(
                f"Sample of type '{type(sample)}` incompatible cannot be converted into an "
                "instance-weighted sample."
            )

    def __len__(self) -> int:
        return len(self.iw)



from .acs import *
from .admissions import *
from .adult import *
from .base import *
from .compas import *
from .credit import *
from .crime import *
from .german import *
from .health import *
from .law import *
from .sqf import *



from dataclasses import dataclass
from typing import Optional
from typing_extensions import override

from conduit.data import TrainValTestSplit
from conduit.data.datamodules.tabular import CdtTabularDataModule
from conduit.data.datasets.tabular import RandomTabularDataset


@dataclass(kw_only=True)
class DummyTabularDataModule(CdtTabularDataModule[RandomTabularDataset]):
    num_samples: int
    num_disc_features: int
    num_cont_features: int
    seed: int = 8
    s_card: Optional[int] = None
    y_card: Optional[int] = None

    @override
    def _get_splits(self) -> TrainValTestSplit[RandomTabularDataset]:
        # Split the data randomly according to val- and test-prop
        data = RandomTabularDataset(
            num_cont_features=self.num_cont_features,
            num_disc_features=self.num_disc_features,
            num_samples=self.num_samples,
            seed=self.seed,
            s_card=self.s_card,
            y_card=self.y_card,
        )
        val_data, test_data, train_data = data.random_split(props=(self.val_prop, self.test_prop))
        return TrainValTestSplit(train=train_data, val=val_data, test=test_data)

    @override
    def _default_transforms(self) -> None:
        return None



from abc import abstractmethod
from dataclasses import dataclass
from typing import TypeVar
from typing_extensions import override

from ranzen import some

from conduit.data.datamodules import CdtDataModule
from conduit.data.datasets.tabular import CdtTabularDataset
from conduit.data.datasets.wrappers import InstanceWeightedDataset
from conduit.data.structures import TernarySample
from conduit.transforms.tabular import TabularNormalize
from conduit.types import Stage

__all__ = ["CdtTabularDataModule"]

D = TypeVar("D", bound=CdtTabularDataset)


@dataclass(kw_only=True, eq=False)
class CdtTabularDataModule(CdtDataModule[D, TernarySample]):
    """Base for tabular data modules."""

    @abstractmethod
    def _default_transforms(self) -> TabularNormalize | None:
        """Return the default tabular transform."""
        raise NotImplementedError()

    @override
    def _setup(self, stage: Stage | None = None) -> None:
        super()._setup(stage)

        if some(tf := self._default_transforms()):
            assert not isinstance(
                self._train_data, InstanceWeightedDataset
            ), "Combination of instance weights and tfs is not yet supported for tabular data."
            assert some(self._train_data)
            assert some(self._val_data)
            assert some(self._test_data)

            self._train_data.fit_transform_(tf)
            self._val_data.transform_(tf)
            self._test_data.transform_(tf)



from pathlib import Path
from typing import List, Literal, Optional, Sequence, Union, overload
from typing_extensions import override

import numpy as np
import numpy.typing as npt
import torch
from torch import Tensor
import torchaudio  # type: ignore

from conduit.data.datasets.base import CdtDataset, I, S, Y
from conduit.data.datasets.utils import (
    AudioLoadingBackend,
    AudioTform,
    apply_audio_transform,
    infer_al_backend,
)
from conduit.data.structures import TargetData
from conduit.types import IndexType

__all__ = ["CdtAudioDataset"]


class CdtAudioDataset(CdtDataset[I, npt.NDArray[np.bytes_], Y, S]):
    """Base dataset for audio data."""

    x: npt.NDArray[np.bytes_]

    def __init__(
        self,
        audio_dir: Union[Path, str],
        *,
        x: npt.NDArray[np.bytes_],
        y: Optional[TargetData] = None,
        s: Optional[TargetData] = None,
        transform: Optional[AudioTform] = None,
    ) -> None:
        super().__init__(x=x, y=y, s=s)

        # Convert string path to Path object.
        if isinstance(audio_dir, str):
            audio_dir = Path(audio_dir)

        self.audio_dir = audio_dir
        self.transform = transform

        # Infer the appropriate audio-loading backend based on the operating system.
        self.al_backend: AudioLoadingBackend = infer_al_backend()
        self.logger.info(f"Using {self.al_backend} as backend for audio-loading.")
        torchaudio.set_audio_backend(self.al_backend)

    def __repr__(self) -> str:
        head = f"Dataset {self.__class__.__name__}"
        body = [
            f"Number of datapoints: {len(self)}",
            f"Base audio-directory location: {self.audio_dir.resolve()}",
            *self.extra_repr().splitlines(),
        ]
        if hasattr(self, "transform") and self.transform is not None:
            body += [repr(self.transform)]
        lines = [head] + [" " * self._repr_indent + line for line in body]
        return '\n'.join(lines)

    def load_sample(self, index: IndexType) -> Tensor:
        def _load(_filename: Path) -> Tensor:
            return torchaudio.load(self.audio_dir / _filename)[0]  # type: ignore

        if isinstance(index, int):
            return _load(self.audio_dir / self.x[index])
        return torch.cat([_load(filepath) for filepath in self.x[index]], dim=0)

    @overload
    def _sample_x(self, index: int, *, coerce_to_tensor: Literal[True]) -> Tensor:
        ...

    @overload
    def _sample_x(
        self, index: Union[List[int], slice], *, coerce_to_tensor: Literal[True]
    ) -> Union[Tensor, Sequence[Tensor]]:
        ...

    @overload
    def _sample_x(self, index: int, *, coerce_to_tensor: Literal[False] = ...) -> Tensor:
        ...

    @overload
    def _sample_x(
        self, index: Union[List[int], slice], *, coerce_to_tensor: Literal[False] = ...
    ) -> Union[Tensor, Sequence[Tensor]]:
        ...

    @override
    def _sample_x(
        self, index: IndexType, *, coerce_to_tensor: bool = False
    ) -> Union[Tensor, Sequence[Tensor]]:
        waveform = self.load_sample(index)
        return apply_audio_transform(waveform, transform=None)



from abc import ABC, abstractmethod
import math
from typing import ClassVar, List, Union, final
from typing_extensions import override

import torch
from torch import Tensor

__all__ = [
    "TabularTransform",
    "MinMaxNormalize",
    "QuantileNormalize",
    "TabularNormalize",
    "ZScoreNormalize",
]


class TabularTransform(ABC):
    """PLaceholder base class."""

    @abstractmethod
    def __call__(self, data: Tensor) -> Tensor:
        raise NotImplementedError()


class TabularNormalize(TabularTransform):
    _EPS: ClassVar[float] = torch.finfo(torch.float32).eps

    def __init__(
        self, inplace: bool = False, indices: Union[slice, List[int]] = slice(None)
    ) -> None:
        self.inplace = inplace
        self.col_indices = indices
        self._is_fitted = False

    @property
    @final
    def is_fitted(self) -> bool:
        return self._is_fitted

    def _maybe_clone(self, data: Tensor) -> Tensor:
        if not self.inplace:
            data = data.clone()
        return data

    @abstractmethod
    def _fit(self, data: Tensor) -> None:
        """inplace operation."""

    @final
    def fit(self, data: Tensor) -> "TabularNormalize":
        self._fit(data[:, self.col_indices])
        self._is_fitted = True
        return self

    @final
    def fit_transform(self, data: Tensor) -> Tensor:
        self.fit(data)
        return self.transform(data)

    @abstractmethod
    def _inverse_transform(self, data: Tensor) -> Tensor:
        """Can be in-place."""

    @final
    def inverse_transform(self, data: Tensor) -> Tensor:
        if not self.is_fitted:
            raise RuntimeError(
                f"Cannot inverse-transform the data with {self.__class__.__name__} because "
                "'fit' has not yet been called."
            )
        data = self._maybe_clone(data)
        data[:, self.col_indices] = self._inverse_transform(data[:, self.col_indices])
        return data

    @abstractmethod
    def _transform(self, data: Tensor) -> Tensor:
        """Can be in-place."""

    @final
    def transform(self, data: Tensor) -> Tensor:
        if not self.is_fitted:
            raise RuntimeError(
                f"Cannot transform the data with {self.__class__.__name__}.fit because "
                "'fit' has not yet been called."
            )
        data = self._maybe_clone(data)
        data[:, self.col_indices] = self._transform(data[:, self.col_indices])
        return data

    @final
    def __call__(self, data: Tensor) -> Tensor:
        return self.transform(data)


@final
class ZScoreNormalize(TabularNormalize):
    mean: Tensor
    std: Tensor

    @override
    def _fit(self, data: Tensor) -> None:
        self.std, self.mean = torch.std_mean(data, dim=0, keepdim=True, unbiased=True)

    @override
    def _inverse_transform(self, data: Tensor) -> Tensor:
        data *= self.std
        data += self.mean
        return data

    @override
    def _transform(self, data: Tensor) -> Tensor:
        data -= self.mean
        data /= self.std.clamp_min(self._EPS)
        return data


@final
class QuantileNormalize(TabularNormalize):
    iqr: Tensor
    median: Tensor

    def __init__(self, q_min: float = 0.25, q_max: float = 0.75, inplace: bool = False) -> None:
        super().__init__(inplace=inplace)
        if not (0 <= q_min <= 1):
            raise ValueError("q_min must be in the range [0, 1]")
        if not (0 <= q_max <= 1):
            raise ValueError("q_max must be in the range [0, 1]")
        if q_min > q_max:
            raise ValueError("'q_min' cannot be greater than 'q_max'.")
        self.q_min = q_min
        self.q_max = q_max

    @staticmethod
    def _compute_quantile(q: float, sorted_values: Tensor) -> Tensor:
        q_ind_frac, q_ind_int = math.modf(q * (len(sorted_values) - 1))
        q_ind_int = int(q_ind_int)
        q_quantile = sorted_values[q_ind_int]
        if q_ind_frac > 0:
            q_quantile += (sorted_values[q_ind_int + 1] - q_quantile) * q_ind_frac
        return q_quantile

    @override
    def _fit(self, data: Tensor) -> None:
        sorted_values = data.sort(dim=0, descending=False).values
        # Compute the 'lower quantile'
        q_min_quantile = self._compute_quantile(q=self.q_min, sorted_values=sorted_values)
        # Compute the 'upper quantile'
        q_max_quantile = self._compute_quantile(q=self.q_max, sorted_values=sorted_values)
        # Compute the interquantile range
        self.iqr = q_max_quantile - q_min_quantile
        self.median = self._compute_quantile(q=0.5, sorted_values=sorted_values)

    @override
    def _inverse_transform(self, data: Tensor) -> Tensor:
        data *= self.iqr
        data += self.median
        return data

    @override
    def _transform(self, data: Tensor) -> Tensor:
        data -= self.median
        data /= self.iqr.clamp_min(self._EPS)
        return data


@final
class MinMaxNormalize(TabularNormalize):
    orig_max: Tensor
    orig_min: Tensor
    orig_range: Tensor

    def __init__(self, new_min: float = 0.0, new_max: float = 1.0, inplace: bool = False) -> None:
        super().__init__(inplace=inplace)
        if new_min > new_max:
            raise ValueError("'new_min' cannot be greater than 'new_max'.")
        self.new_min = new_min
        self.new_max = new_max
        self.new_range = self.new_max - self.new_min

    @override
    def _fit(self, data: Tensor) -> None:
        self.orig_min = torch.min(data, dim=0, keepdim=True).values
        self.orig_max = torch.max(data, dim=0, keepdim=True).values
        self.orig_range = self.orig_max - self.orig_min

    @override
    def _inverse_transform(self, data: Tensor) -> Tensor:
        data -= self.new_min
        data /= self.new_range + self._EPS
        data *= self.orig_range
        data += self.orig_min
        return data

    @override
    def _transform(self, data: Tensor) -> Tensor:
        data -= self.orig_min
        data /= self.orig_range.clamp_min(self._EPS)
        data *= self.new_range
        data += self.new_min
        return data



from enum import auto
from pathlib import Path
from typing import ClassVar, Optional, Union
from typing_extensions import TypeAlias

import pandas as pd
from ranzen import StrEnum
import torch
from torch import Tensor

from conduit.data.datasets.utils import UrlFileInfo, download_from_url
from conduit.data.structures import TernarySample

from .base import CdtVisionDataset
from .utils import ImageTform

__all__ = ["FMoW", "FMoWSplit", "FMoWSplitScheme"]


class FMoWSplit(StrEnum):
    TRAIN = "train"
    VAL = "val"
    ID_VAL = "id_val"
    TEST = "test"
    ID_TEST = "id_test"


class FMoWSplitScheme(StrEnum):
    # Since 3 years are set aside for validation, the earliest year that can be used for splitting
    # -- such that the training set is not empty -- is 2003 + 3 = 2006.
    POST_2006 = auto()
    POST_2007 = auto()
    POST_2008 = auto()
    POST_2009 = auto()
    POST_2010 = auto()
    POST_2011 = auto()
    POST_2012 = auto()
    POST_2013 = auto()
    POST_2014 = auto()
    POST_2015 = auto()
    POST_2016 = auto()
    POST_2017 = auto()
    POST_2018 = auto()
    OFFICIAL = POST_2018
    "Official split: equivalent to 'POST_2018'"


SampleType: TypeAlias = TernarySample


class FMoW(CdtVisionDataset[TernarySample, Tensor, Tensor]):
    """The functional Map of the World (fMoW; land use / building classification) dataset.
    This is a port of the iWildCam dataset from `WILDS`_ which defines a preprocessing procedure for
    the dataset introduced in `fMoW`_.

    Per `fMoW`_: We present a new dataset, Functional Map of the World (fMoW), which aims to
    inspire the development of machine learning models capable of predicting the functional purpose
    of buildings and land use ['y' label] from temporal sequences of satellite images and a rich set of metadata
    features. The metadata provided with each image enables reasoning about location, time, sun
    angles, physical sizes, and other features when making predictions about objects in the image.
    Our dataset consists of over 1 million images from over 200 countries. For each image, we
    provide at least one bounding box annotation containing one of 63 categories, including a "false
    detection" category.

    For the sake of the WILDS becnhmark 'region' is defined as the grouping attribute on which the
    evaluation metrics are conditioned. This also holds here as we seek to keep everything
    functionally equivalent with respect to the WILDS version of the dataset.

    .. _WILDS:
        https://arxiv.org/abs/1711.07846

    .. _fMoW:
        https://arxiv.org/abs/1711.07846
    """

    SampleType: TypeAlias = TernarySample
    Split: TypeAlias = FMoWSplit
    SplitScheme: TypeAlias = FMoWSplitScheme

    _NUM_VAL_YEARS: ClassVar[int] = 3
    _BASE_DIR_NAME: ClassVar[str] = "fmow"
    _IMG_DIR_NAME: ClassVar[str] = "images"
    _METADATA_FILENAME: ClassVar[str] = "rgb_metadata.csv"
    _CC_MAPPER_FILENAME: ClassVar[str] = "country_code_mapping.csv"

    _FILE_INFO: ClassVar[UrlFileInfo] = UrlFileInfo(
        name="fmow.tar.gz",
        url="https://worksheets.codalab.org/rest/bundles/0xaec91eb7c9d548ebb15e1b5e60f966ab/contents/blob/",
        md5=None,
    )

    def __init__(
        self,
        root: Union[str, Path],
        *,
        split: Optional[Union[FMoWSplit, str]] = None,
        split_scheme: Optional[Union[FMoWSplitScheme, str]] = FMoWSplitScheme.OFFICIAL,
        drop_other: bool = True,
        transform: Optional[ImageTform] = None,
        download: bool = True,
    ) -> None:
        """
        :param root: Root directory of the dataset.
        :param split: Which predefined split of the dataset to use. If ``None`` then the full
            (unsplit) dataset will be returned.
        :param transform: A function/transform that takes in a PIL or ndarray image and returns a
            transformed version. E.g, ``transforms.RandomCrop``.
        :param split_scheme: Which splitting scheme to use. The task is treated as a time-series
            one: the data is split temporally such that the training data always precedes the test
            data in time. This is ignored if ``split`` is ``None`` (the entire dataset is used).
        :param drop_other: Whether to drop samples belonging to uncategorised ('Other') regions.
            Such samples make up a very small portion of the dataset and so we drop these by default
            given that they should be (according to WILDS) excluded from evaluation.
        :param download: If ``True``, downloads the dataset from the internet and puts it in the
            root directory. If the dataset is already downloaded, it is not downloaded again.
        :raises FileNotFoundError: If ``download=False`` and an existing dataset cannot be found in
            the root directory.
        """
        self.split = FMoWSplit(split) if isinstance(split, str) else split
        self.split_scheme = (
            FMoWSplitScheme(split_scheme) if isinstance(split_scheme, str) else split_scheme
        )
        self.root = Path(root)
        self._base_dir = self.root / self._BASE_DIR_NAME
        self._img_dir = self._base_dir / self._IMG_DIR_NAME
        self.download = download
        if not self._check_unzipped():
            if self.download:
                download_from_url(
                    file_info=self._FILE_INFO,
                    root=self.root,
                    logger=self.logger,
                    remove_finished=True,
                )

            else:
                raise FileNotFoundError(
                    f"Data not found at location {self._base_dir.resolve()}. Have you"
                    " downloaded it?"
                )

        # Read in metadata
        self.metadata = pd.read_csv(self._base_dir / self._METADATA_FILENAME)
        # filter out sequestered images
        self.metadata = self.metadata.loc[self.metadata['split'] != 'seq']
        country_codes = pd.read_csv(self._base_dir / self._CC_MAPPER_FILENAME)
        region_mapper = dict(zip(country_codes['alpha-3'], country_codes["region"]))
        # Map the country codes to regions using the supplementary file.
        self.metadata["region"] = self.metadata.country_code.map(region_mapper)
        # In the WILDS code the 'Other'  region is excluded from evaluation; since there are so few
        # samples sourced from 'Other' regions we instead take the tack of dropping these samples
        # by default.
        if drop_other:
            self.metadata.dropna(subset="region")
        else:
            self.metadata.fillna("Other", inplace=True)
        # label encode the class (land use category) labels -- we do this prior splitting to
        # ensure consistency in the mappings.
        self.metadata["category_le"] = self.metadata["category"].factorize()[0]
        # label encode the subgroup (reegion) labels -- we do this prior to splitting to ensure
        # consistency in the mappings.
        self.metadata["region_le"] = self.metadata["region"].factorize()[0]
        # extract the years from the timestamps.
        self.metadata["year"] = self.metadata.timestamp.str.split("-", expand=True)[0].astype(int)

        # Use a split of the data if one is specified, else just use all of the data
        if self.split is not None:
            split_mask = self.metadata.split == self.split.value
            self.metadata = self.metadata.loc[split_mask]

            if self.split_scheme is not None:
                test_threshold = int(self.split_scheme.value.split('_')[-1])
                # OOD test split
                if self.split is FMoWSplit.TEST:
                    scheme_mask = self.metadata.year >= test_threshold
                # OOD validation split
                elif self.split is FMoWSplit.VAL:
                    # # set aside the final years of the training set for validation
                    scheme_mask = (self.metadata.year < test_threshold) > (
                        self.metadata.year >= (test_threshold - self._NUM_VAL_YEARS)
                    )
                # ID (train/val/test) splits
                else:
                    scheme_mask = self.metadata.year < (test_threshold - self._NUM_VAL_YEARS)
                self.metadata = self.metadata.loc[scheme_mask]

        # Amend the filenames, which are actually based on the sample indexes.
        def index_to_fn(index: int) -> str:
            return f"rgb_img_{index}.png"

        self.metadata["img_filename"] = self.metadata.index.map(index_to_fn)
        x = self.metadata['img_filename'].to_numpy()
        y = torch.as_tensor(self.metadata["category_le"].to_numpy(), dtype=torch.long)
        s = torch.as_tensor(self.metadata["region_le"].to_numpy(), dtype=torch.long)

        super().__init__(x=x, y=y, s=s, transform=transform, image_dir=self._img_dir)

    def _check_unzipped(self) -> bool:
        return (self._img_dir).is_dir()



"""Law Admissions Dataset."""

from dataclasses import dataclass

from ethicml.data import Dataset, Law
from ethicml.data import LawSplits as LawSens

from conduit.fair.data.datamodules.tabular.base import EthicMlDataModule

__all__ = ["LawDataModule", "LawSens"]


@dataclass(kw_only=True)
class LawDataModule(EthicMlDataModule):
    """LSAC Law Admissions Dataset."""

    sens_feat: LawSens = LawSens.SEX
    disc_feats_only: bool = False

    @property
    def em_dataset(self) -> Dataset:
        return Law(split=self.sens_feat, discrete_only=self.disc_feats_only, invert_s=self.invert_s)



"""SSRP Dataset."""

from enum import Enum
from pathlib import Path
from typing import ClassVar, List, Optional, Union, cast
from typing_extensions import TypeAlias

import pandas as pd
import torch
from torch import Tensor

from conduit.data.datasets.utils import GdriveFileInfo, download_from_gdrive
from conduit.data.structures import TernarySample

from .base import CdtVisionDataset
from .utils import ImageTform

__all__ = ["SSRP", "SSRPSplit"]


class SSRPSplit(Enum):
    TASK = "Task"
    PRETRAIN = "Pre_Train"


class SSRP(CdtVisionDataset[TernarySample, Tensor, Tensor]):
    SampleType: TypeAlias = TernarySample
    Split: TypeAlias = SSRPSplit

    _FILE_INFO: ClassVar[GdriveFileInfo] = GdriveFileInfo(
        name="ghaziabad.zip", id="1RE4srtC63VnyU0e1qx16QNdjyyQXg2hj"
    )

    def __init__(
        self,
        root: Union[str, Path],
        *,
        split: Union[SSRPSplit, str] = SSRPSplit.PRETRAIN,
        download: bool = True,
        transform: Optional[ImageTform] = None,
    ) -> None:
        self.root = Path(root)
        self._base_dir = self.root / self.__class__.__name__
        self._metadata_path = self._base_dir / "metadata.csv"
        self.download = download
        self.split = SSRPSplit(split)

        if self.download:
            download_from_gdrive(file_info=self._FILE_INFO, root=self._base_dir, logger=self.logger)
        if not self._check_unzipped():
            raise FileNotFoundError(
                f"Data not found at location {self._base_dir.resolve()}. Have you downloaded it?"
            )
        if not self._metadata_path.exists():
            self._extract_metadata()

        self.metadata = pd.read_csv(self._base_dir / "metadata.csv")
        self.metadata = self.metadata.loc[self.metadata["split"].to_numpy() == self.split.value]

        x = self.metadata["filepath"].to_numpy()
        y = torch.as_tensor(self.metadata["class_le"].to_numpy(), dtype=torch.long)
        s = torch.as_tensor(self.metadata["season_le"].to_numpy(), dtype=torch.long)

        super().__init__(x=x, y=y, s=s, transform=transform, image_dir=self._base_dir)

    def _check_unzipped(self) -> bool:
        return (self._base_dir / "Ghaziabad").is_dir()

    def _extract_metadata(self) -> None:
        """Extract concept/context/superclass information from the image filepaths and it save to csv."""
        self.logger.info("Extracting metadata.")
        image_paths: List[Path] = []
        for ext in ("jpg", "jpeg", "png"):
            # Glob images from child folders recusrively, excluding hidden files
            image_paths.extend(self._base_dir.glob(f"**/[!.]*.{ext}"))
        image_paths_str = [str(image.relative_to(self._base_dir)) for image in image_paths]
        filepaths = pd.Series(image_paths_str)
        metadata = cast(
            pd.DataFrame,
            filepaths.str.split("/", expand=True)
            .dropna(axis=1)
            .rename(columns={0: "region", 1: "split", 2: "class", 3: "filename"}),
        )
        # Extract the seasonal metadata from the filenames
        metadata["season"] = metadata["filename"].str.split(r"\((.*?)\s.*", expand=True)[1]
        metadata["filepath"] = filepaths
        metadata.sort_index(axis=1, inplace=True)
        metadata.sort_values(by=["filepath"], axis=0, inplace=True)
        metadata = self._label_encode_metadata(metadata)
        metadata.to_csv(self._metadata_path)

    @staticmethod
    def _label_encode_metadata(metadata: pd.DataFrame) -> pd.DataFrame:
        """Label encode the extracted concept/context/superclass information."""
        for col in metadata.columns:
            # Skip over filepath and filename columns
            if "file" not in col:
                # Add a new column containing the label-encoded data
                metadata[f"{col}_le"] = metadata[col].factorize()[0]
        return metadata



"""Tabular data-module."""

from abc import abstractmethod
from dataclasses import dataclass, field
from typing import Dict, List, Optional, Union, cast, final
from typing_extensions import override

import ethicml as em
from ethicml.data import Dataset, FeatureOrder
from sklearn.preprocessing import StandardScaler
from torch import Tensor

from conduit.data.datamodules import CdtDataModule
from conduit.data.structures import TernarySample, TrainValTestSplit
from conduit.fair.data.datasets import DataTupleDataset

__all__ = ["EthicMlDataModule"]


@dataclass(kw_only=True)
class EthicMlDataModule(CdtDataModule[DataTupleDataset, TernarySample[Tensor]]):
    """Base data-module for tabular datasets."""

    scaler: em.ScalerType = field(default_factory=StandardScaler)  # pyright: ignore
    invert_s: bool = False
    _datatuple: Optional[em.DataTuple] = field(default=None, init=False)
    _train_datatuple: Optional[em.DataTuple] = field(default=None, init=False)
    _val_datatuple: Optional[em.DataTuple] = field(default=None, init=False)
    _test_datatuple: Optional[em.DataTuple] = field(default=None, init=False)
    _cont_features: Optional[List[str]] = field(default=None, init=False)
    _disc_features: Optional[List[str]] = field(default=None, init=False)
    _feature_groups: Optional[Dict[str, Optional[List[slice]]]] = field(default=None, init=False)

    @property
    @final
    def datatuple(self) -> em.DataTuple:
        self._check_setup_called("datatuple")
        return cast(em.DataTuple, self._datatuple)

    @property
    @abstractmethod
    def em_dataset(self) -> Dataset:
        ...

    @staticmethod
    def _get_split_sizes(train_len: int, *, test_prop: Union[int, float]) -> List[int]:
        """Computes split sizes for train and validation sets."""
        if isinstance(test_prop, int):
            train_len -= test_prop
            splits = [train_len, test_prop]
        elif isinstance(test_prop, float):
            test_len = int(test_prop * train_len)
            train_len -= test_len
            splits = [train_len, test_len]
        else:
            raise ValueError(f"Unsupported type {type(test_prop)}")

        return splits

    @override
    def prepare_data(self) -> None:
        self.make_feature_groups()

    @override
    def _get_splits(self) -> TrainValTestSplit[DataTupleDataset]:
        self._datatuple = self.em_dataset.load(order=FeatureOrder.disc_first)

        data_len = int(self._datatuple.x.shape[0])
        num_train_val, num_test = self._get_split_sizes(data_len, test_prop=self.test_prop)
        train_val, test_data = em.train_test_split(
            data=self._datatuple,
            train_percentage=(1 - (num_test / data_len)),
            random_seed=self.seed,
        )
        _, num_val = self._get_split_sizes(num_train_val, test_prop=self.val_prop)
        train_data, val_data = em.train_test_split(
            data=train_val,
            train_percentage=(1 - (num_val / num_train_val)),
            random_seed=self.seed,
        )

        self._train_datatuple, self.scaler = em.scale_continuous(
            self.em_dataset, datatuple=train_data, scaler=self.scaler
        )
        self._val_datatuple, _ = em.scale_continuous(
            self.em_dataset, datatuple=val_data, scaler=self.scaler, fit=False
        )
        self._test_datatuple, _ = em.scale_continuous(
            self.em_dataset, datatuple=test_data, scaler=self.scaler, fit=False
        )

        train_data = DataTupleDataset(
            dataset=self._train_datatuple,
            disc_features=self.em_dataset.discrete_features,
            cont_features=self.em_dataset.continuous_features,
        )

        val_data = DataTupleDataset(
            dataset=self._val_datatuple,
            disc_features=self.em_dataset.discrete_features,
            cont_features=self.em_dataset.continuous_features,
        )

        test_data = DataTupleDataset(
            dataset=self._test_datatuple,
            disc_features=self.em_dataset.discrete_features,
            cont_features=self.em_dataset.continuous_features,
        )
        return TrainValTestSplit(train=train_data, val=val_data, test=test_data)

    @property
    def train_datatuple(self) -> em.DataTuple:
        assert self._train_datatuple is not None
        return self._train_datatuple

    @property
    def val_datatuple(self) -> em.DataTuple:
        assert self._val_datatuple is not None
        return self._val_datatuple

    @property
    def test_datatuple(self) -> em.DataTuple:
        assert self._test_datatuple is not None
        return self._test_datatuple

    @property
    def feature_groups(self) -> Dict[str, Optional[List[slice]]]:
        assert self._feature_groups is not None
        return self._feature_groups

    @property
    def disc_features(self) -> List[str]:
        assert self._disc_features is not None
        return self._disc_features

    @property
    def cont_features(self) -> List[str]:
        assert self._cont_features is not None
        return self._cont_features

    def make_feature_groups(self) -> None:
        """Make feature groups for reconstruction."""
        self._disc_features = self.em_dataset.discrete_features
        self._cont_features = self.em_dataset.continuous_features
        assert self.em_dataset.disc_feature_groups is not None
        self._feature_groups = dict(
            discrete=self.grouped_features_indexes(self.em_dataset.disc_feature_groups)
        )

    @staticmethod
    def grouped_features_indexes(group_iter: Dict[str, List[str]]) -> List[slice]:
        """Group discrete features names according to the first segment of their name.

        Then return a list of their corresponding slices (assumes order is maintained).
        """

        feature_slices = []
        start_idx = 0
        for group in group_iter.values():
            len_group = len(list(group))
            indexes = slice(start_idx, start_idx + len_group)
            feature_slices.append(indexes)
            start_idx += len_group

        return feature_slices



# Generated by configen, do not edit.
# See https://github.com/facebookresearch/hydra/tree/master/tools/configen
# fmt: off
# isort:skip_file
# flake8: noqa

from dataclasses import dataclass, field
from conduit.data.datasets.audio.ecoacoustics import SoundscapeAttr
from omegaconf import MISSING
from ranzen.torch.data import TrainingMode
from typing import Any
from typing import List
from typing import Optional


@dataclass
class EcoacousticsDataModuleConf:
    _target_: str = "conduit.data.datamodules.audio.EcoacousticsDataModule"
    root: str = MISSING
    train_batch_size: int = 64
    eval_batch_size: Optional[int] = None
    val_prop: float = 0.2
    test_prop: float = 0.2
    num_workers: int = 0
    seed: int = 47
    persist_workers: bool = False
    pin_memory: bool = True
    stratified_sampling: bool = False
    instance_weighting: bool = False
    training_mode: TrainingMode = TrainingMode.epoch
    train_tf: Any = None  # Optional[Callable[[Tensor], Tensor]]
    test_tf: Any = None  # Optional[Callable[[Tensor], Tensor]]
    segment_len: float = 15
    sample_rate: int = 48000
    target_attrs: List[SoundscapeAttr] = MISSING



import logging
from typing import (
    ClassVar,
    Generic,
    List,
    Literal,
    Optional,
    Sequence,
    TypeVar,
    Union,
    cast,
    final,
    overload,
)
from typing_extensions import Self, override

import numpy as np
import numpy.typing as npt
from ranzen.misc import gcopy
import torch
from torch import Tensor

from conduit.data.structures import (
    BinarySample,
    LoadedData,
    NamedSample,
    SampleBase,
    SizedDataset,
    SubgroupSample,
    TargetData,
    TernarySample,
    UnloadedData,
)
from conduit.logging import init_logger
from conduit.metrics import merge_indices
from conduit.types import IndexType

__all__ = ["CdtDataset", "I", "S", "X", "Y"]

X = TypeVar("X", bound=UnloadedData)
S = TypeVar("S", bound=Optional[Tensor])
Y = TypeVar("Y", bound=Optional[Tensor])
I = TypeVar("I", bound=SampleBase, covariant=True)


class CdtDataset(SizedDataset, Generic[I, X, Y, S]):
    _repr_indent: ClassVar[int] = 4
    _logger: Optional[logging.Logger] = None

    def __init__(
        self, *, x: X, y: Optional[TargetData] = None, s: Optional[TargetData] = None
    ) -> None:
        self.x = x
        if isinstance(y, np.ndarray):
            y = torch.as_tensor(y)
        if isinstance(s, np.ndarray):
            s = torch.as_tensor(s)
        self.y = cast(Y, y if y is None else y.squeeze())
        self.s = cast(S, s if s is None else s.squeeze())

        self._dim_x: Optional[torch.Size] = None
        self._dim_s: Optional[torch.Size] = None
        self._dim_y: Optional[torch.Size] = None
        self._card_y: Optional[int] = None
        self._card_s: Optional[int] = None

    def __repr__(self) -> str:
        head = f"Dataset {self.__class__.__name__}"
        body = [f"Number of datapoints: {len(self)}"]
        body += self.extra_repr().splitlines()
        lines = [head] + [" " * self._repr_indent + line for line in body]
        return '\n'.join(lines)

    @staticmethod
    def extra_repr() -> str:
        return ""

    @property
    def logger(self) -> logging.Logger:
        if self._logger is None:
            self._logger = init_logger(self.__class__.__name__)
        return self._logger

    @overload
    def _sample_x(self, index: int, *, coerce_to_tensor: Literal[True]) -> Tensor:
        ...

    @overload
    def _sample_x(
        self, index: Union[List[int], slice], *, coerce_to_tensor: Literal[True]
    ) -> Union[Tensor, Sequence[Tensor]]:
        ...

    @overload
    def _sample_x(self, index: int, *, coerce_to_tensor: Literal[False] = ...) -> LoadedData:
        ...

    @overload
    def _sample_x(
        self, index: Union[List[int], slice], *, coerce_to_tensor: Literal[False] = ...
    ) -> Union[LoadedData, Sequence[LoadedData]]:
        ...

    def _sample_x(
        self, index: IndexType, *, coerce_to_tensor: bool = False
    ) -> Union[LoadedData, Tensor, Sequence[LoadedData]]:
        x = self.x[index]
        if coerce_to_tensor and (not isinstance(x, Tensor)):
            x = torch.as_tensor(x)
        return x

    def _sample_s(self, index: IndexType) -> Optional[Tensor]:
        return None if self.s is None else self.s[index]

    def _sample_y(self, index: IndexType) -> Optional[Tensor]:
        return None if self.y is None else self.y[index]

    @property
    @final
    def dim_x(
        self,
    ) -> torch.Size:
        if self._dim_x is None:
            self._dim_x = self._sample_x(0, coerce_to_tensor=True).shape
        return self._dim_x

    @property
    @final
    def dim_s(
        self,
    ) -> torch.Size:
        if self.s is None:
            cls_name = self.__class__.__name__
            raise AttributeError(
                f"'{cls_name}.dim_s' cannot be determined as '{cls_name}.s' is 'None'"
            )
        if self._dim_s is None:
            self._dim_s = torch.Size((1,)) if self.s.ndim == 1 else self.s.shape[1:]
        return self._dim_s

    @property
    @final
    def dim_y(
        self,
    ) -> torch.Size:
        if self.y is None:
            cls_name = self.__class__.__name__
            raise AttributeError(
                f"'{cls_name}.dim_y' cannot be determined as '{cls_name}.y' is 'None'"
            )
        elif self._dim_y is None:
            self._dim_y = torch.Size((1,)) if self.y.ndim == 1 else self.y.shape[1:]
        return self._dim_y

    @property
    @final
    def card_y(
        self,
    ) -> int:
        if self.y is None:
            cls_name = self.__class__.__name__
            raise AttributeError(
                f"'{cls_name}.card_y' cannot be determined as '{cls_name}.y' is 'None'"
            )
        if self._card_y is None:
            self._card_y = len(self.y.unique())
        return self._card_y

    @property
    @final
    def card_s(
        self,
    ) -> int:
        if self.s is None:
            cls_name = self.__class__.__name__
            raise AttributeError(
                f"'{cls_name}.card_s' cannot be determined as '{cls_name}.s' is 'None'"
            )
        if self._card_s is None:
            self._card_s = len(self.s.unique())
        return self._card_s

    @property
    @final
    def groups(
        self,
    ) -> Tensor:
        valid_s = (self.s is not None) and (self.s.dtype is torch.long)
        valid_y = (self.y is not None) and (self.y.dtype is torch.long)
        if not valid_y and not valid_s:
            raise AttributeError(
                "Unable to compute group ids for dataset because 'y' and 's' are both 'None' or are"
                "not of dtype `torch.long`."
            )
        indices = []
        # 'merge_indices' takes the last element of the 'indices' argument as the seed
        # for the merged index set -- for consistency we use the 'y' label as the seed
        # such that the binary version of merge_indices produces the bijective mapping
        # g <- y * card(s) + s.
        if valid_s:
            indices.append(cast(Tensor, self.s))
        if valid_y:
            indices.append(cast(Tensor, self.y))
        return merge_indices(*indices, return_cardinalities=False)

    def subset(
        self: Self,
        indices: Union[List[int], npt.NDArray[np.uint64], Tensor, slice],
        *,
        deep: bool = False,
    ) -> Self:
        """Create a subset of the dataset from the given indices.

        :param indices: The sample-indices from which to create the subset. In the case of being a
            numpy array or tensor, said array or tensor must be 0- or 1-dimensional.

        :param deep: Whether to create a copy of the underlying dataset as a basis for the subset.
            If False then the data of the subset will be a view of original dataset's data.

        :returns: A subset of the dataset from the given indices.
        """
        # lazily import make_subset to prevent it being a circular import
        from conduit.data.datasets.utils import make_subset

        return make_subset(dataset=self, indices=indices, deep=deep)

    def random_split(
        self: Self,
        props: Union[Sequence[float], float],
        *,
        deep: bool = False,
        seed: Optional[int] = None,
        reproducible: bool = False,
    ) -> List[Self]:
        """Randomly split the dataset into subsets according to the given proportions.

        :param props: The fractional size of each subset into which to randomly split the data.
            Elements must be non-negative and sum to 1 or less; if less then the size of the final
            split will be computed by complement.

        :param deep: Whether to create a copy of the underlying dataset as a basis for the random
            subsets. If False then the data of the subsets will be views of original dataset's data.

        :param seed: PRNG seed to use for splitting the data.

        :param reproducible: Whether to make the split reproducible.

        :returns: Random subsets of the data of the requested proportions.
        """
        # lazily import ``random_split`` to prevent it from being a circular import
        from conduit.data.datasets.utils import random_split

        return random_split(
            self, props=props, deep=deep, seed=seed, as_indices=False, reproducible=reproducible
        )

    @overload
    def cat(self, other: Self, *, inplace: Literal[True], deep: bool = ...) -> None:
        ...

    @overload
    def cat(self, other: Self, *, inplace: Literal[False] = ..., deep: bool = ...) -> Self:
        ...

    @overload
    def cat(self, other: Self, *, inplace: bool = ..., deep: bool = ...) -> Optional[Self]:
        ...

    def cat(self, other: Self, *, inplace: bool = False, deep: bool = False) -> Optional[Self]:
        """Concatenate this ``self`` with another dataset of the same type.

        :param other: Other dataset to concatenate with this instance
        :param inplace: Whether to concatenate in place.
        :param deep: Whether to create a deep copy of this dataset as the basis for the superset.

        :returns: A concatenation of ``self`` with ``other``.

        .. note::
            All data-independent attributes will be inherited from ``self``.
        """
        superset = self if inplace else gcopy(self, deep=deep)
        xs = [superset.x, other.x]
        if isinstance(superset.x, np.ndarray):
            superset.x = np.concatenate(xs, axis=0)
        else:
            superset.x = torch.cat(cast(List[Tensor], xs), dim=0)
        if (superset.s is not None) and (other.s is not None):
            superset.s = torch.cat([superset.s, other.s], dim=0)
        if (superset.y is not None) and (other.y is not None):
            superset.y = torch.cat([superset.y, other.y], dim=0)

        if not inplace:
            return superset

    @override
    def __getitem__(self: Self, index: int) -> I:
        x = self._sample_x(index, coerce_to_tensor=False)
        y = self._sample_y(index)
        s = self._sample_s(index)
        # Fetch the appropriate 'Sample' class
        if y is None:
            sample = NamedSample(x=x) if s is None else SubgroupSample(x=x, s=s)
        elif s is None:
            sample = BinarySample(x=x, y=y)
        else:
            sample = TernarySample(x=x, y=y, s=s)
        return sample  # type: ignore

    @override
    def __len__(self) -> int:
        return len(self.x)

    def __iadd__(self, other: Self) -> Self:
        self.cat(other, inplace=True, deep=False)
        return self

    def __add__(self, other: Self) -> Self:
        return self.cat(other, inplace=False, deep=False)



"""COMPAS Dataset."""

from dataclasses import dataclass

from ethicml.data import Compas, Dataset
from ethicml.data import CompasSplits as CompasSens

from conduit.fair.data.datamodules.tabular.base import EthicMlDataModule

__all__ = ["CompasDataModule", "CompasSens"]


@dataclass(kw_only=True)
class CompasDataModule(EthicMlDataModule):
    """COMPAS Dataset."""

    sens_feat: CompasSens = CompasSens.SEX
    disc_feats_only: bool = False

    @property
    def em_dataset(self) -> Dataset:
        return Compas(
            split=self.sens_feat, discrete_only=self.disc_feats_only, invert_s=self.invert_s
        )



from dataclasses import dataclass
from typing import Callable, Generic, List, Optional, TypeVar, Union
from typing_extensions import Self, override

from PIL import Image
import numpy as np
from ranzen.misc import gcopy
from torch import Tensor

from conduit.data.datasets.vision.utils import ImageTform, apply_image_transform
from conduit.data.structures import InputContainer, RawImage, concatenate_inputs

__all__ = [
    "FixMatchPair",
    "FixMatchTransform",
]

X = TypeVar("X", bound=Union[Tensor, RawImage, List[Image.Image]])


@dataclass
class FixMatchPair(InputContainer, Generic[X]):
    strong: X
    weak: X

    @override
    def __len__(self) -> int:
        if isinstance(self.strong, Image.Image):
            return 1
        return len(self.strong)

    @override
    def __add__(self, other: Self) -> Self:
        copy = gcopy(self, deep=False)
        is_batched = isinstance(self.strong, (Tensor, np.ndarray)) and (self.strong.ndim == 4)
        copy.strong = concatenate_inputs(copy.strong, other.strong, is_batched=is_batched)
        copy.weak = concatenate_inputs(copy.weak, other.weak, is_batched=is_batched)

        return copy


A = TypeVar("A", bound=ImageTform)


class FixMatchTransform(Generic[A]):
    def __init__(
        self,
        strong_transform: A,
        *,
        weak_transform: A,
        shared_transform_start: Optional[Callable[[Union[RawImage, Tensor]], RawImage]] = None,
        shared_transform_end: Optional[
            Callable[[Union[RawImage, Tensor]], Union[RawImage, Tensor]]
        ] = None,
    ) -> None:
        self.strong_transform = strong_transform
        self.weak_transform = weak_transform
        self.shared_transform_start = shared_transform_start
        self.shared_transform_end = shared_transform_end

    def __call__(self, image: RawImage) -> FixMatchPair:
        if self.shared_transform_start is not None:
            image = self.shared_transform_start(image)

        strongly_augmented_image = apply_image_transform(
            image=image, transform=self.strong_transform
        )

        weakly_augmented_image = apply_image_transform(image=image, transform=self.weak_transform)
        if self.shared_transform_end is not None:
            strongly_augmented_image = self.shared_transform_end(strongly_augmented_image)
            weakly_augmented_image = self.shared_transform_end(weakly_augmented_image)

        return FixMatchPair(strong=strongly_augmented_image, weak=weakly_augmented_image)



from enum import Enum
from pathlib import Path
from typing import ClassVar, Optional, Union, cast
from typing_extensions import TypeAlias

from numpy import typing as npt
import pandas as pd
from ranzen import str_to_enum
from sklearn.preprocessing import MultiLabelBinarizer
import torch
from torch import Tensor

from conduit.data.structures import TernarySample

from .base import CdtVisionDataset
from .utils import ImageTform

__all__ = ["NIHChestXRays", "NIHSplit", "NIHSubgroup", "NIHTarget"]


class NIHSplit(Enum):
    TRAIN = "train_val_list.txt"
    TEST = "test_list.txt"


class NIHSubgroup(Enum):
    GENDER = "Patient Gender"
    AGE = "Patient Age"


class NIHTarget(Enum):
    """
    Taget attributes for the NIH chest X-rays dataset.

    Fraction of labels that are positive for each of the thoracic diseases:
        Atelectasis           0.103095
        Cardiomegaly          0.024759
        Consolidation         0.041625
        Edema                 0.020540
        Effusion              0.118775
        Emphysema             0.022440
        Fibrosis              0.015037
        Hernia                0.002025
        Infiltration          0.177435
        Mass                  0.051570
        Nodule                0.056466
        Pleural_Thickening    0.030191
        Pneumonia             0.012763
        Pneumothorax          0.04728
    """

    ATELECTASIS = "Atelectasis"
    CARDIOMEGALY = "Cardiomegaly"
    CONSOLIDATION = "Consolidation"
    EDEMA = "Edema"
    EFFUSION = "Effusion"
    EMPHYSEMA = "Emphysema"
    FIBROSIS = "Fibrosis"
    HERNIA = "Hernia"
    INFILTRATION = "Infiltration"
    MASS = "Mass"
    NODULE = "Nodule"
    PLEURAL_THICKENING = "Pleural_Thickening"
    PNEUMONIA = "Pneumonia"
    PNEUMOTHORAX = "Pneumothorax"
    FINDING = "No Finding"


class NIHChestXRays(CdtVisionDataset[TernarySample, Tensor, Tensor]):
    """ "
    National Institutes of Health (NIH) chest X-Rays dataset.
    This NIH Chest X-rays dataset is comprised of 112,120 X-ray images with disease labels from
    30,805 unique patients. To create these labels, the authors used Natural Language Processing to
    text-mine disease classifications from the associated radiological reports. The labels are
    expected to be >90% accurate and suitable for weakly-supervised learning. The original radiology
    reports are not publicly available but you can find more details on the labeling process in
    `this
    <https://www.nih.gov/news-events/news-releases/nih-clinical-center-provides-one-largest-publicly-available-chest-x-ray-datasets-scientific-community>`__
    Open Access paper.

    The dataset can be downloaded by following the above link or from `kaggle <https://www.kaggle.com/datasets/nih-chest-xrays/data>`__
    """

    SampleType: TypeAlias = TernarySample
    Target: TypeAlias = NIHTarget
    Subgroup: TypeAlias = NIHSubgroup
    Split: TypeAlias = NIHSplit

    _METADATA_FILENAME: ClassVar[str] = "Data_Entry_2017.csv"
    _BASE_DIR_NAME: ClassVar[str] = "nih_chest_x_rays"

    def __init__(
        self,
        root: Union[Path, str],
        *,
        target: Optional[Union[NIHTarget, str]] = NIHTarget.FINDING,
        subgroup: Union[NIHSubgroup, str] = NIHSubgroup.GENDER,
        split: Optional[Union[NIHSplit, str]] = None,
        transform: Optional[ImageTform] = None,
        num_quantiles: Optional[int] = 4,
        download: bool = True,
    ) -> None:
        """
        :param root: Root directory of the dataset.
        :param target: Attribute to set as the target attribute ('y').
        :param subgroup: Attribute to set as the subgroup attribute ('s').
        :param split: Which predefined split of the dataset to use. If ``None`` then the full
            (unsplit) dataset will be returned.
        :param transform: A function/transform that takes in a PIL or ndarray image and returns a
            transformed version. E.g, ``transforms.RandomCrop``.
        :param num_quantiles: Number of quantiles (equal-sized buckets) to bin 'Patient Age' into
            (only applicable when ``sens_attr=NIHSensAttr.AGE``). E.g. 10 for deciles, 4 for
            quartiles.
        :param download: If ``True``, downloads the dataset from the internet and puts it in the
            root directory. If the dataset is already downloaded, it is not downloaded again.
            This requires kaggle credentials.
        :raises FileNotFoundError: If ``download=False`` and an existing dataset cannot be found in
            the root directory.
        """
        self.root = Path(root)
        self._base_dir = self.root / self._BASE_DIR_NAME
        self._metadata_fp = self._base_dir / self._METADATA_FILENAME
        if target is not None:
            target = str_to_enum(str_=target, enum=NIHTarget)
        self.target = target
        self.subgroup = str_to_enum(str_=subgroup, enum=NIHSubgroup)
        if split is not None:
            split = str_to_enum(str_=split, enum=NIHSplit)
        self.split = split
        self.download = download
        if not self._check_unzipped():
            if self.download:
                import kaggle  # type: ignore

                kaggle.api.authenticate()
                kaggle.api.dataset_download_files(
                    dataset="nih-chest-xrays/data",
                    path=self._base_dir,
                    unzip=True,
                    quiet=False,
                )
            else:
                raise FileNotFoundError(
                    f"Data not found at location {self.root.resolve()}. Have you downloaded it?"
                )

        self.metadata = cast(pd.DataFrame, pd.read_csv(self._metadata_fp))
        if self.split is not None:
            split_info = pd.read_csv(
                self._base_dir / self.split.value, header=None, names=["Image Index"]
            )
            self.metadata = self.metadata.merge(split_info, on="Image Index")

        # In the case of Patient Gender, factorize yields the mapping: M -> 0, F -> 1
        s_pd = self.metadata[self.subgroup.value]
        if (self.subgroup is NIHSubgroup.AGE) and (num_quantiles is not None):
            s_pd = cast(pd.Series, pd.qcut(s_pd, q=num_quantiles))
        s_pd_le = s_pd.factorize()[0]
        s = torch.as_tensor(s_pd_le, dtype=torch.long)

        findings_str = self.metadata["Finding Labels"].str.split("|")
        self.encoder = MultiLabelBinarizer().fit(findings_str)
        findings_ml = pd.DataFrame(
            cast(npt.NDArray, self.encoder.transform(findings_str)), columns=self.encoder.classes_
        )
        self.metadata = pd.concat((self.metadata, findings_ml), axis=1)
        if self.target is None:
            findings_ml.drop("No Finding", axis=1, inplace=True)
        else:
            findings_ml = findings_ml[self.target.value]
            if self.target is NIHTarget.FINDING:
                # Flip the label such that the presence of some thoracic disease
                # is represented by the positive class.
                findings_ml = 1 - findings_ml
        y = torch.as_tensor(findings_ml.to_numpy(), dtype=torch.long)

        filepaths = (
            pd.Series(list(self._base_dir.glob("*/*/*")), dtype=str)
            .sort_values()
            .str.split("/", expand=True, n=4)[4]
            .rename("Image Path")
        )
        if self.split is None:
            self.metadata = pd.concat((filepaths, self.metadata), axis=1)
        else:
            filenames = filepaths.str.rsplit("/", expand=True, n=1)[1].rename("Image Index")
            filepaths = pd.concat((filepaths, filenames), axis=1)
            self.metadata = self.metadata.merge(filepaths, on="Image Index")
        x = self.metadata["Image Path"].to_numpy()
        super().__init__(image_dir=self._base_dir, x=x, s=s, y=y, transform=transform)

    def _check_unzipped(self) -> bool:
        return (self._metadata_fp).exists()



"""ISIC Dataset."""

from enum import auto
from itertools import islice
import os
from pathlib import Path
import shutil
from typing import TYPE_CHECKING, ClassVar, Iterable, Iterator, List, Optional, TypeVar, Union
from typing_extensions import TypeAlias
import zipfile

from PIL import Image
import pandas as pd
from ranzen import StrEnum, flatten_dict
import requests
import torch
from torch import Tensor
from tqdm import tqdm

from conduit.data.structures import TernarySample

from .base import CdtVisionDataset
from .utils import ImageTform

if TYPE_CHECKING:
    from _typeshed import SupportsRead

__all__ = ["IsicAttr", "ISIC"]


class IsicAttr(StrEnum):
    HISTO = auto()
    MALIGNANT = auto()
    PATCH = auto()


T = TypeVar("T")
SampleType: TypeAlias = TernarySample


class ISIC(CdtVisionDataset[SampleType, Tensor, Tensor]):
    """PyTorch Dataset for the ISIC 2018 dataset from
    'Skin Lesion Analysis Toward Melanoma Detection 2018: A Challenge Hosted by the International
    Skin Imaging Collaboration (ISIC)',"""

    SampleType: TypeAlias = TernarySample
    Attr: TypeAlias = IsicAttr

    LABELS_FILENAME: ClassVar[str] = "labels.csv"
    METADATA_FILENAME: ClassVar[str] = "metadata.csv"
    _PBAR_COL: ClassVar[str] = "#fac000"
    _REST_API_URL: ClassVar[str] = "https://isic-archive.com/api/v1"

    def __init__(
        self,
        root: Union[str, Path],
        *,
        download: bool = True,
        max_samples: int = 25_000,  # default is the number of samples used for the NSLB paper
        target_attr: IsicAttr = IsicAttr.MALIGNANT,
        context_attr: IsicAttr = IsicAttr.HISTO,
        transform: Optional[ImageTform] = None,
    ) -> None:
        self.target_attr = IsicAttr(target_attr) if isinstance(target_attr, str) else target_attr
        self.context_attr = (
            IsicAttr(context_attr) if isinstance(context_attr, str) else context_attr
        )

        self.root = Path(root)
        self.download = download
        self._base_dir = self.root / self.__class__.__name__
        self._processed_dir = self._base_dir / "processed"
        self._raw_dir = self._base_dir / "raw"

        if max_samples < 1:
            raise ValueError("max_samples must be a positive integer.")
        self.max_samples = max_samples
        if self.download:
            self._download_data()
            self._preprocess_data()
        elif not self._check_downloaded():
            raise FileNotFoundError(
                f"Data not found at location {self._processed_dir.resolve()}. "
                "Have you downloaded it?"
            )

        self.metadata = pd.read_csv(self._processed_dir / self.LABELS_FILENAME)
        # Divide up the dataframe into its constituent arrays because indexing with pandas is
        # considerably slower than indexing with numpy/torch
        x = self.metadata["path"].to_numpy()
        s = torch.as_tensor(self.metadata[str(self.context_attr)], dtype=torch.int32)
        y = torch.as_tensor(self.metadata[str(self.target_attr)], dtype=torch.int32)

        super().__init__(x=x, y=y, s=s, transform=transform, image_dir=self._processed_dir)

    def _check_downloaded(self) -> bool:
        return (self._raw_dir / "images").exists() and (
            self._raw_dir / self.METADATA_FILENAME
        ).exists()

    def _check_processed(self) -> bool:
        return (self._processed_dir / "ISIC-images").exists() and (
            self._processed_dir / self.LABELS_FILENAME
        ).exists()

    @staticmethod
    def chunk(it: Iterable[T], *, size: int) -> Iterator[List[T]]:
        """Divide any iterable into chunks of the given size."""
        it = iter(it)
        return iter(lambda: list(islice(it, size)), [])  # this is magic from stackoverflow

    def _download_isic_metadata(self) -> pd.DataFrame:
        """Downloads the metadata CSV from the ISIC website."""
        self._raw_dir.mkdir(parents=True, exist_ok=True)
        req = requests.get(
            f"{self._REST_API_URL}/image?limit={self.max_samples}&sort=name&sortdir=1&detail=false"
        )
        image_ids = req.json()
        image_ids = [image_id["_id"] for image_id in image_ids]

        template_start = "?limit=300&sort=name&sortdir=1&detail=true&imageIds=%5B%22"
        template_sep = "%22%2C%22"
        template_end = "%22%5D"
        entries = []
        with tqdm(
            total=(len(image_ids) - 1) // 300 + 1,
            desc="Downloading metadata",
            colour=self._PBAR_COL,
        ) as pbar:
            for block in self.chunk(image_ids, size=300):
                pbar.set_postfix(image_id=block[0])
                args = ""
                args += template_start
                args += template_sep.join(block)
                args += template_end
                req = requests.get(f"{self._REST_API_URL}/image{args}")
                image_details = req.json()
                for image_detail in image_details:
                    entry = flatten_dict(image_detail, sep=".")
                    entries.append(entry)
                pbar.update()

        metadata_df = pd.DataFrame(entries)
        metadata_df = metadata_df.set_index("_id")
        metadata_df.to_csv(self._raw_dir / self.METADATA_FILENAME)
        return metadata_df

    def _download_isic_images(self) -> None:
        """Given the metadata CSV, downloads the ISIC images."""
        metadata_path = self._raw_dir / self.METADATA_FILENAME
        if not metadata_path.is_file():
            raise FileNotFoundError(
                f"{self.METADATA_FILENAME} not downloaded. "
                "Run 'download_isic_data` before this function."
            )
        metadata_df = pd.read_csv(metadata_path)
        metadata_df = metadata_df.set_index("_id")

        template_start = "?include=images&imageIds=%5B%22"
        template_sep = "%22%2C%22"
        template_end = "%22%5D"
        raw_image_dir = self._raw_dir / "images"
        raw_image_dir.mkdir(exist_ok=True)
        image_ids = list(metadata_df.index)
        with tqdm(
            total=(len(image_ids) - 1) // 50 + 1, desc="Downloading images", colour=self._PBAR_COL
        ) as pbar:
            for i, block in enumerate(self.chunk(image_ids, size=50)):
                pbar.set_postfix(image_id=block[0])
                args = ""
                args += template_start
                args += template_sep.join(block)
                args += template_end
                req = requests.get(f"{self._REST_API_URL}/image/download{args}", stream=True)
                req.raise_for_status()
                image_path = raw_image_dir / f"{i}.zip"
                with image_path.open("wb") as f:
                    src: SupportsRead[bytes] = req.raw  # Pyright needs help with type inference...
                    shutil.copyfileobj(src, f)
                del req
                pbar.update()

    def _preprocess_isic_metadata(self) -> None:
        """Preprocesses the raw ISIC metadata."""
        self._processed_dir.mkdir(exist_ok=True)

        metadata_path = self._raw_dir / self.METADATA_FILENAME
        if not metadata_path.is_file():
            raise FileNotFoundError(
                f"{self.METADATA_FILENAME} not found while preprocessing ISIC dataset. "
                "Run `download_isic_metadata` and `download_isic_images` before "
                "calling `preprocess_isic_metadata`."
            )
        metadata_df = pd.read_csv(metadata_path)
        metadata_df = metadata_df.set_index("_id")
        labels_df = self._remove_uncertain_diagnoses(metadata_df)
        labels_df = self._add_patch_column(labels_df)

        labels_df["path"] = (
            str(self._processed_dir)
            + os.sep
            + "ISIC-images"
            + os.sep
            + labels_df["dataset.name"]
            + os.sep
            + labels_df["name"]
            + ".jpg"
        )
        labels_df.to_csv(self._processed_dir / self.LABELS_FILENAME)

    def _preprocess_isic_images(self) -> None:
        """Preprocesses the images."""
        if (self._processed_dir / "ISIC-images").is_dir():
            return
        if not (self._raw_dir / "images").is_dir():
            raise FileNotFoundError(
                "Raw ISIC images not found. Run `download_isic_images` before "
                "calling `preprocess_isic_images`."
            )
        labels_df = pd.read_csv(self._processed_dir / self.LABELS_FILENAME)
        labels_df = labels_df.set_index("_id")

        self._processed_dir.mkdir(exist_ok=True)
        image_zips = tuple((self._raw_dir / "images").glob("**/*.zip"))
        with tqdm(total=len(image_zips), desc="Unzipping images", colour=self._PBAR_COL) as pbar:
            for file in image_zips:
                pbar.set_postfix(file_index=file.stem)
                with zipfile.ZipFile(file, "r") as zip_ref:
                    zip_ref.extractall(self._processed_dir)
                    pbar.update()
        images: List[Path] = []
        for ext in ("jpg", "jpeg", "png", "gif"):
            images.extend(self._processed_dir.glob(f"**/*.{ext}"))
        with tqdm(total=len(images), desc="Processing images", colour=self._PBAR_COL) as pbar:
            for image_path in images:
                pbar.set_postfix(image_name=image_path.stem)
                image = Image.open(image_path)
                image = image.resize((224, 224))  # Resize the images to be of size 224 x 224
                if image.mode in ("RGBA", "P"):
                    image = image.convert("RGB")
                image.save(str(image_path.rename(image_path.with_suffix(".jpg"))))
                pbar.update()

    @staticmethod
    def _remove_uncertain_diagnoses(metadata_df: pd.DataFrame) -> pd.DataFrame:
        labels_df = metadata_df.loc[
            metadata_df["meta.clinical.benign_malignant"].isin({"benign", "malignant"})
        ]  # throw out unknowns
        malignant_mask = labels_df["meta.clinical.benign_malignant"] == "malignant"
        labels_df["malignant"] = malignant_mask.astype("uint8")

        labels_df["meta.clinical.diagnosis_confirm_type"].fillna(
            value="non-histopathology", inplace=True
        )
        histopathology_mask = labels_df["meta.clinical.diagnosis_confirm_type"] == "histopathology"
        labels_df["histo"] = histopathology_mask.astype("uint8")

        return labels_df

    @staticmethod
    def _add_patch_column(labels_df: pd.DataFrame) -> pd.DataFrame:
        """Adds a patch column to the input DataFrame."""
        patch_mask = labels_df["dataset.name"] == "SONIC"
        # add to labels_df
        labels_df["patch"] = None
        labels_df.loc[patch_mask, "patch"] = 1
        labels_df.loc[~patch_mask, "patch"] = 0
        assert all(patch is not None for patch in labels_df["patch"])
        return labels_df

    def _download_data(self) -> None:
        """Attempt to download data if files cannot be found in the base folder."""
        # # Check whether the data has already been downloaded - if it has and the integrity
        # # of the files can be confirmed, then we are done
        if self._check_downloaded():
            self.logger.info("Files already downloaded and verified.")
            return
        # Create the directory and any required ancestors if not already existent
        self._base_dir.mkdir(exist_ok=True, parents=True)
        self.logger.info(f"Downloading metadata into {self._raw_dir / self.METADATA_FILENAME}...")
        self._download_isic_metadata()
        self.logger.info(
            f"Downloading data into {self._raw_dir} for up to {self.max_samples} samples..."
        )
        self._download_isic_images()

    def _preprocess_data(self) -> None:
        """Preprocess the downloaded data if the processed image-directory/metadata don't exist."""
        # If the data has already been processed, skip this operation
        if self._check_processed():
            self.logger.info("Metadata and images already preprocessed.")
            return
        self.logger.info(
            "Preprocessing metadata (adding columns, removing uncertain diagnoses) and saving into"
            f" {str(self._processed_dir / self.LABELS_FILENAME)}..."
        )
        self._preprocess_isic_metadata()
        self.logger.info(
            "Preprocessing images (transforming to 3-channel RGB, resizing to 224x224) and saving "
            f"into{str(self._processed_dir / 'ISIC-images')}..."
        )
        self._preprocess_isic_images()



"""NICO++ data-module."""

from dataclasses import dataclass
from typing import Any, List, Optional
from typing_extensions import override

import albumentations as A  # type: ignore

from conduit.data.datamodules.vision.base import CdtVisionDataModule
from conduit.data.datasets.vision import NICOPP, NicoPPSplit, NicoPPTarget, SampleType
from conduit.data.structures import TrainValTestSplit

__all__ = ["NICOPPDataModule"]


@dataclass(kw_only=True)
class NICOPPDataModule(CdtVisionDataModule[SampleType]):
    """Data-module for the NICO dataset."""

    image_size: int = 224
    superclasses: Optional[List[NicoPPTarget]] = None

    @property
    @override
    def _default_train_transforms(self) -> A.Compose:
        base_transforms = A.Compose(
            [
                A.Resize(self.image_size, self.image_size),
                A.CenterCrop(self.image_size, self.image_size),
            ]
        )
        normalization = super()._default_train_transforms
        return A.Compose([base_transforms, normalization])

    @property
    @override
    def _default_test_transforms(self) -> A.Compose:
        return self._default_train_transforms

    @override
    def prepare_data(self, *args: Any, **kwargs: Any) -> None:
        pass

    @override
    def _get_image_splits(self) -> TrainValTestSplit[NICOPP]:
        train_data, val_data, test_data = (
            NICOPP(root=self.root, superclasses=self.superclasses, transform=None, split=split)
            for split in NicoPPSplit
        )

        return TrainValTestSplit(train=train_data, val=val_data, test=test_data)



"""ColoredMNIST data-module."""

from dataclasses import dataclass
from functools import partial
from typing import Any, Dict, List, Optional
from typing_extensions import override

import albumentations as A  # type: ignore
from albumentations.pytorch.transforms import ToTensorV2  # type: ignore
import numpy as np
from torchvision.datasets import MNIST  # type: ignore

from conduit.data.datamodules.vision.base import CdtVisionDataModule
from conduit.data.datasets.vision.cmnist import ColoredMNIST, ColoredMNISTSplit, SampleType
from conduit.data.structures import MeanStd, TrainValTestSplit

__all__ = ["ColoredMNISTDataModule"]


@dataclass(kw_only=True)
class ColoredMNISTDataModule(CdtVisionDataModule[SampleType]):
    """Data-module for the ColoredMNIST dataset."""

    image_size: int = 32
    use_predefined_splits: bool = False
    # Colorization settings
    label_map: Optional[Dict[int, int]] = None
    colors: Optional[List[int]] = None
    num_colors: int = 10
    scale: float = 0.2
    correlation: float = 1.0
    binarize: bool = False
    greyscale: bool = False
    background: bool = False
    black: bool = True

    @property
    @override
    def _default_train_transforms(self) -> A.Compose:
        transforms = [A.Compose([A.Resize(self.image_size, self.image_size)])]
        if self.norm_values is not None:
            normalization = A.Compose(
                [A.Normalize(mean=self.norm_values.mean, std=self.norm_values.std), ToTensorV2()]
            )
            transforms.append(normalization)
        return A.Compose(transforms)

    @property
    @override
    def _default_test_transforms(self) -> A.Compose:
        return self._default_train_transforms

    @override
    def prepare_data(self, *args: Any, **kwargs: Any) -> None:
        MNIST(root=str(self.root), download=True, train=True)
        MNIST(root=str(self.root), download=True, train=False)

    @override
    def _get_image_splits(self) -> TrainValTestSplit[ColoredMNIST]:
        # TODO: Add more sophisticated (e.g. biased) splits
        fact_func = partial(
            ColoredMNIST,
            root=self.root,
            background=self.background,
            black=self.black,
            greyscale=self.greyscale,
            correlation=self.correlation,
            colors=self.colors,
            num_colors=self.num_colors,
            label_map=self.label_map,
        )
        # Use the predefined train and test splits, sampling the val split
        # randomly from the train split
        if self.use_predefined_splits:
            train_data = fact_func(split=ColoredMNISTSplit.TRAIN)
            test_data = fact_func(split=ColoredMNISTSplit.TEST)
            val_data, train_data = train_data.random_split(props=self.val_prop)
        else:
            # Split the data randomly according to val- and test-prop
            train_data = fact_func(split=None)
            val_data, test_data, train_data = train_data.random_split(
                props=(self.val_prop, self.test_prop), seed=self.seed
            )
        # Compute the channel-wise first and second moments
        channel_means = np.mean(train_data.x, axis=(0, 1, 2)) / 255.0
        channel_stds = np.std(train_data.x, axis=(0, 1, 2)) / 255.0

        self.norm_values = MeanStd(mean=channel_means.tolist(), std=channel_stds.tolist())

        return TrainValTestSplit(train=train_data, val=val_data, test=test_data)



# Generated by configen, do not edit.
# See https://github.com/facebookresearch/hydra/tree/master/tools/configen
# fmt: off
# isort:skip_file
# flake8: noqa

from dataclasses import dataclass, field
from conduit.data.datasets.audio.ecoacoustics import SoundscapeAttr
from omegaconf import MISSING
from typing import Any
from typing import List


@dataclass
class EcoacousticsConf:
    _target_: str = "conduit.data.datasets.audio.Ecoacoustics"
    root: str = MISSING
    target_attrs: List[SoundscapeAttr] = MISSING
    transform: Any = None  # Optional[Callable[[Tensor], Tensor]]
    download: bool = True
    segment_len: float = 15
    sample_rate: int = 48000



"""Adult Income Dataset."""

from dataclasses import dataclass

from ethicml.data import Adult, Dataset
from ethicml.data import AdultSplits as AdultSens

from conduit.fair.data.datamodules.tabular.base import EthicMlDataModule

__all__ = ["AdultDataModule", "AdultSens"]


@dataclass(kw_only=True)
class AdultDataModule(EthicMlDataModule):
    """UCI Adult Income Dataset."""

    bin_nationality: bool = False
    sens_feat: AdultSens = AdultSens.SEX
    bin_race: bool = False
    disc_feats_only: bool = False

    @property
    def em_dataset(self) -> Dataset:
        return Adult(
            split=self.sens_feat,
            binarize_nationality=self.bin_nationality,
            discrete_only=self.disc_feats_only,
            binarize_race=self.bin_race,
            invert_s=self.invert_s,
        )



from dataclasses import dataclass, field
from typing_extensions import override

from conduit.data import TrainValTestSplit
from conduit.data.datamodules.tabular import CdtTabularDataModule
from conduit.fair.data.datasets.acs import (
    ACSDataset,
    ACSHorizon,
    ACSSetting,
    ACSState,
    ACSSurvey,
    ACSSurveyYear,
)
from conduit.transforms import TabularNormalize, ZScoreNormalize

__all__ = ["ACSDataModule"]


@dataclass(kw_only=True)
class ACSDataModule(CdtTabularDataModule[ACSDataset]):
    setting: ACSSetting
    survey_year: ACSSurveyYear = ACSSurveyYear.YEAR_2018
    horizon: ACSHorizon = ACSHorizon.ONE_YEAR
    survey: ACSSurvey = ACSSurvey.PERSON
    states: list[ACSState] = field(default_factory=lambda: [ACSState.AL])
    class_train_props: dict[int, dict[int, float]] | None = None

    @override
    def _get_splits(self) -> TrainValTestSplit[ACSDataset]:
        data = ACSDataset(
            setting=self.setting,
            survey_year=self.survey_year,
            horizon=self.horizon,
            survey=self.survey,
            states=self.states,
        )
        train_data, val_data, test_data = data.subsampled_split(
            train_props=self.class_train_props,
            val_prop=self.val_prop,
            test_prop=self.test_prop,
            seed=self.seed,
        )
        return TrainValTestSplit(train=train_data, val=val_data, test=test_data)

    @override
    def _default_transforms(self) -> TabularNormalize:
        return ZScoreNormalize()



"""Utils for vision datasets."""

from pathlib import Path
from typing import Any, Callable, Literal, Optional, Union, get_args, overload
from typing_extensions import TypeAlias

from PIL import Image
import albumentations as A  # type: ignore
import cv2
import numpy as np
import numpy.typing as npt
import torch
from torch import Tensor
from torchvision.transforms import functional as TF  # type: ignore

from conduit.data.structures import RawImage

__all__ = [
    "AlbumentationsTform",
    "ImageLoadingBackend",
    "ImageTform",
    "PillowTform",
    "apply_image_transform",
    "img_to_tensor",
    "infer_il_backend",
    "load_image",
]


ImageLoadingBackend: TypeAlias = Literal["opencv", "pillow"]


@overload
def load_image(
    filepath: Union[Path, str], *, backend: Literal["opencv"] = ...
) -> npt.NDArray[np.integer]:
    ...


@overload
def load_image(filepath: Union[Path, str], *, backend: Literal["pillow"]) -> Image.Image:
    ...


def load_image(filepath: Union[Path, str], *, backend: ImageLoadingBackend = "opencv") -> RawImage:
    """Load an image from disk using the requested backend.

    :param filepath: The path of the image-file to be loaded.
    :param backend: Backed to use for loading the image: either 'opencv' or 'pillow'.

    :returns: The loaded image file as a numpy array if 'opencv' was the selected backend
        and a PIL image otherwise.
    :raises OSError: if the images can't be read.
    """
    if backend == "opencv":
        if isinstance(filepath, Path):
            # cv2 can only read string filepaths
            filepath = str(filepath)
        image = cv2.imread(filepath)
        if image is None:
            raise OSError(f"Image-file could not be read from location '{filepath}'")
        return cv2.cvtColor(image, cv2.COLOR_BGR2RGB)  # type: ignore
    return Image.open(filepath)


AlbumentationsTform: TypeAlias = Union[A.Compose, A.BasicTransform]
PillowTform: TypeAlias = Callable[[Image.Image], Any]
ImageTform: TypeAlias = Union[AlbumentationsTform, PillowTform]


def infer_il_backend(transform: Optional[ImageTform]) -> ImageLoadingBackend:
    """Infer which image-loading backend to use based on the type of the image-transform.

    :param transform: The image transform from which to infer the image-loading backend.
        If the transform is derived from Albumentations, then 'opencv' will be selected as the
        backend, else 'pillow' will be selected.

    :returns: The backend to load images with based on the supplied image-transform: either
        'opencv' or 'pillow'.
    """
    # Default to openccv is transform is None as numpy arrays are generally
    # more tractable
    if transform is None or isinstance(transform, get_args(AlbumentationsTform)):
        return "opencv"
    return "pillow"


def apply_image_transform(
    image: RawImage, *, transform: Optional[ImageTform]
) -> Union[RawImage, Tensor]:
    image_ = image
    if transform is not None:
        if isinstance(transform, (A.Compose, A.BasicTransform)):
            if isinstance(image, Image.Image):
                image = np.array(image)
            image_ = transform(image=image)["image"]
        else:
            if isinstance(image, np.ndarray):
                image = Image.fromarray(image)
            image_ = transform(image)
    return image_


def img_to_tensor(img: Union[Image.Image, np.ndarray]) -> Tensor:
    if isinstance(img, Image.Image):
        return TF.pil_to_tensor(img)
    return torch.from_numpy(
        np.moveaxis(img / (255.0 if img.dtype == np.uint8 else 1), -1, 0).astype(np.float32)
    )



from typing import List, Optional

import numpy as np
import pandas as pd
import torch

from conduit.data.datasets.tabular import CdtTabularDataset

__all__ = ["RandomTabularDataset"]


class RandomTabularDataset(CdtTabularDataset):
    def __init__(
        self,
        num_disc_features: int = 3,
        num_cont_features: int = 5,
        num_samples: int = 256,
        seed: int = 0,
        s_card: Optional[int] = None,
        y_card: Optional[int] = None,
    ) -> None:
        rng = np.random.default_rng(seed)
        feats_dict = {}
        feature_groups: List[slice] = []
        prev = 0
        for i in range(num_disc_features):
            num_classes = int(rng.integers(low=2, high=10, size=1)[0])
            disc_feat = pd.Series(rng.integers(low=0, high=num_classes, size=num_samples))
            ohe_feats = pd.get_dummies(disc_feat)
            feats_dict[f"disc_{i}"] = ohe_feats
            feature_groups += [slice(prev, prev + ohe_feats.shape[1])]
            prev += ohe_feats.shape[1]

        feats = pd.concat(feats_dict.values(), axis=1)
        num_ohe_feats = feats.shape[1]

        for i in range(num_cont_features):
            cont_feat = rng.random(num_samples)
            feats[f"cont_{i}"] = cont_feat

        if y_card is None:
            y = None
        else:
            y = torch.as_tensor(rng.integers(low=0, high=y_card, size=num_samples))

        if s_card is None:
            s = None
        else:
            s = torch.as_tensor(rng.integers(low=0, high=s_card, size=num_samples))

        super().__init__(
            x=torch.as_tensor(feats.to_numpy(dtype=np.float32)),
            y=y,
            s=s,
            feature_groups=feature_groups,
            non_ohe_indexes=list(range(num_ohe_feats, feats.shape[1])),
        )



from typing import Final

from conduit.data.structures import MeanStd

__all__ = ["IMAGENET_STATS"]

IMAGENET_STATS: Final[MeanStd] = MeanStd(
    mean=(0.485, 0.456, 0.406),
    std=(0.229, 0.224, 0.225),
)



# Generated by configen, do not edit.
# See https://github.com/facebookresearch/hydra/tree/master/tools/configen
# fmt: off
# isort:skip_file
# flake8: noqa

from dataclasses import dataclass, field
from ethicml.data.tabular_data.admissions import AdmissionsSplits
from ethicml.data.tabular_data.adult import AdultSplits
from ethicml.data.tabular_data.compas import CompasSplits
from ethicml.data.tabular_data.credit import CreditSplits
from ethicml.data.tabular_data.crime import CrimeSplits
from ethicml.data.tabular_data.german import GermanSplits
from ethicml.data.tabular_data.health import HealthSplits
from ethicml.data.tabular_data.law import LawSplits
from ethicml.data.tabular_data.sqf import SqfSplits
from omegaconf import MISSING
from ranzen.torch.data import TrainingMode
from typing import Any
from typing import Optional


@dataclass
class AdmissionsDataModuleConf:
    _target_: str = "conduit.fair.data.datamodules.AdmissionsDataModule"
    train_batch_size: int = 64
    eval_batch_size: Optional[int] = None
    val_prop: float = 0.2
    test_prop: float = 0.2
    num_workers: int = 0
    seed: int = 47
    persist_workers: bool = False
    pin_memory: bool = True
    stratified_sampling: bool = False
    instance_weighting: bool = False
    training_mode: TrainingMode = TrainingMode.epoch
    scaler: Any = MISSING  # ScalerType
    invert_s: bool = False
    sens_feat: AdmissionsSplits = AdmissionsSplits.GENDER
    disc_feats_only: bool = False


@dataclass
class AdultDataModuleConf:
    _target_: str = "conduit.fair.data.datamodules.AdultDataModule"
    train_batch_size: int = 64
    eval_batch_size: Optional[int] = None
    val_prop: float = 0.2
    test_prop: float = 0.2
    num_workers: int = 0
    seed: int = 47
    persist_workers: bool = False
    pin_memory: bool = True
    stratified_sampling: bool = False
    instance_weighting: bool = False
    training_mode: TrainingMode = TrainingMode.epoch
    scaler: Any = MISSING  # ScalerType
    invert_s: bool = False
    bin_nationality: bool = False
    sens_feat: AdultSplits = AdultSplits.SEX
    bin_race: bool = False
    disc_feats_only: bool = False


@dataclass
class CompasDataModuleConf:
    _target_: str = "conduit.fair.data.datamodules.CompasDataModule"
    train_batch_size: int = 64
    eval_batch_size: Optional[int] = None
    val_prop: float = 0.2
    test_prop: float = 0.2
    num_workers: int = 0
    seed: int = 47
    persist_workers: bool = False
    pin_memory: bool = True
    stratified_sampling: bool = False
    instance_weighting: bool = False
    training_mode: TrainingMode = TrainingMode.epoch
    scaler: Any = MISSING  # ScalerType
    invert_s: bool = False
    sens_feat: CompasSplits = CompasSplits.SEX
    disc_feats_only: bool = False


@dataclass
class CreditDataModuleConf:
    _target_: str = "conduit.fair.data.datamodules.CreditDataModule"
    train_batch_size: int = 64
    eval_batch_size: Optional[int] = None
    val_prop: float = 0.2
    test_prop: float = 0.2
    num_workers: int = 0
    seed: int = 47
    persist_workers: bool = False
    pin_memory: bool = True
    stratified_sampling: bool = False
    instance_weighting: bool = False
    training_mode: TrainingMode = TrainingMode.epoch
    scaler: Any = MISSING  # ScalerType
    invert_s: bool = False
    sens_feat: CreditSplits = CreditSplits.SEX
    disc_feats_only: bool = False


@dataclass
class CrimeDataModuleConf:
    _target_: str = "conduit.fair.data.datamodules.CrimeDataModule"
    train_batch_size: int = 64
    eval_batch_size: Optional[int] = None
    val_prop: float = 0.2
    test_prop: float = 0.2
    num_workers: int = 0
    seed: int = 47
    persist_workers: bool = False
    pin_memory: bool = True
    stratified_sampling: bool = False
    instance_weighting: bool = False
    training_mode: TrainingMode = TrainingMode.epoch
    scaler: Any = MISSING  # ScalerType
    invert_s: bool = False
    sens_feat: CrimeSplits = CrimeSplits.RACE_BINARY
    disc_feats_only: bool = False


@dataclass
class GermanDataModuleConf:
    _target_: str = "conduit.fair.data.datamodules.GermanDataModule"
    train_batch_size: int = 64
    eval_batch_size: Optional[int] = None
    val_prop: float = 0.2
    test_prop: float = 0.2
    num_workers: int = 0
    seed: int = 47
    persist_workers: bool = False
    pin_memory: bool = True
    stratified_sampling: bool = False
    instance_weighting: bool = False
    training_mode: TrainingMode = TrainingMode.epoch
    scaler: Any = MISSING  # ScalerType
    invert_s: bool = False
    sens_feat: GermanSplits = GermanSplits.SEX
    disc_feats_only: bool = False


@dataclass
class HealthDataModuleConf:
    _target_: str = "conduit.fair.data.datamodules.HealthDataModule"
    train_batch_size: int = 64
    eval_batch_size: Optional[int] = None
    val_prop: float = 0.2
    test_prop: float = 0.2
    num_workers: int = 0
    seed: int = 47
    persist_workers: bool = False
    pin_memory: bool = True
    stratified_sampling: bool = False
    instance_weighting: bool = False
    training_mode: TrainingMode = TrainingMode.epoch
    scaler: Any = MISSING  # ScalerType
    invert_s: bool = False
    sens_feat: HealthSplits = HealthSplits.SEX
    disc_feats_only: bool = False


@dataclass
class LawDataModuleConf:
    _target_: str = "conduit.fair.data.datamodules.LawDataModule"
    train_batch_size: int = 64
    eval_batch_size: Optional[int] = None
    val_prop: float = 0.2
    test_prop: float = 0.2
    num_workers: int = 0
    seed: int = 47
    persist_workers: bool = False
    pin_memory: bool = True
    stratified_sampling: bool = False
    instance_weighting: bool = False
    training_mode: TrainingMode = TrainingMode.epoch
    scaler: Any = MISSING  # ScalerType
    invert_s: bool = False
    sens_feat: LawSplits = LawSplits.SEX
    disc_feats_only: bool = False


@dataclass
class SqfDataModuleConf:
    _target_: str = "conduit.fair.data.datamodules.SqfDataModule"
    train_batch_size: int = 64
    eval_batch_size: Optional[int] = None
    val_prop: float = 0.2
    test_prop: float = 0.2
    num_workers: int = 0
    seed: int = 47
    persist_workers: bool = False
    pin_memory: bool = True
    stratified_sampling: bool = False
    instance_weighting: bool = False
    training_mode: TrainingMode = TrainingMode.epoch
    scaler: Any = MISSING  # ScalerType
    invert_s: bool = False
    sens_feat: SqfSplits = SqfSplits.SEX
    disc_feats_only: bool = False



"""NICO Dataset."""

from enum import auto
from pathlib import Path
from typing import ClassVar, List, Optional, Union, cast
from typing_extensions import TypeAlias

from PIL import Image, UnidentifiedImageError
import pandas as pd
from ranzen import StrEnum
import torch
from torch import Tensor

from conduit.data.datasets.utils import GdriveFileInfo, download_from_gdrive
from conduit.data.structures import TernarySample

from .base import CdtVisionDataset
from .utils import ImageTform

__all__ = [
    "NICO",
    "NicoSuperclass",
]


class NicoSuperclass(StrEnum):
    ANIMALS = auto()
    VEHICLES = auto()


SampleType: TypeAlias = TernarySample


class NICO(CdtVisionDataset[TernarySample, Tensor, Tensor]):
    """Datset for Non-I.I.D. image classification introduced in
    'Towards Non-I.I.D. Image Classification: A Dataset and Baselines'
    """

    SampleType: TypeAlias = TernarySample
    Superclass: TypeAlias = NicoSuperclass

    _FILE_INFO: ClassVar[GdriveFileInfo] = GdriveFileInfo(
        name="NICO.zip",
        id="1L6cHNhuwwvrolukBklFyhFu7Y8WUUIQ7",
        md5="78c686f84e31ad6b6c052f97ed5f532b",
    )

    def __init__(
        self,
        root: Union[str, Path],
        *,
        download: bool = True,
        transform: Optional[ImageTform] = None,
        superclass: Optional[Union[NicoSuperclass, str]] = NicoSuperclass.ANIMALS,
    ) -> None:
        self.superclass = NicoSuperclass(superclass) if isinstance(superclass, str) else superclass
        self.root = Path(root)
        self.download = download
        self._base_dir = self.root / self.__class__.__name__
        self._metadata_path = self._base_dir / "metadata.csv"

        if self.download:
            download_from_gdrive(file_info=self._FILE_INFO, root=self.root, logger=self.logger)
        elif not self._check_unzipped():
            raise FileNotFoundError(
                f"Data not found at location {self._base_dir.resolve()}. Have you downloaded it?"
            )
        if not self._metadata_path.exists():
            self._extract_metadata()

        self.metadata = pd.read_csv(self._base_dir / "metadata.csv")
        self.class_tree = (
            self.metadata[["concept", "context"]]
            .drop_duplicates()
            .groupby("concept")
            .agg(set)
            .to_dict()["context"]
        )
        self.concept_label_decoder = (
            self.metadata[["concept", "concept_le"]].set_index("concept_le").to_dict()["concept"]
        )
        self.context_label_decoder = (
            self.metadata[["context", "context_le"]].set_index("context_le").to_dict()["context"]
        )

        if self.superclass is not None:
            self.metadata = self.metadata[self.metadata["superclass"] == str(self.superclass)]
        # # Divide up the dataframe into its constituent arrays because indexing with pandas is
        # # substantially slower than indexing with numpy/torch
        x = self.metadata["filepath"].to_numpy()
        y = torch.as_tensor(self.metadata["concept_le"].to_numpy(), dtype=torch.long)
        s = torch.as_tensor(self.metadata["context_le"].to_numpy(), dtype=torch.long)

        super().__init__(x=x, y=y, s=s, transform=transform, image_dir=self._base_dir)

    def _check_unzipped(self) -> bool:
        return all((self._base_dir / sc.name).exists() for sc in NicoSuperclass)

    def _extract_metadata(self) -> None:
        """Extract concept/context/superclass information from the image filepaths and it save to csv."""
        self.logger.info("Extracting metadata.")
        image_paths: List[Path] = []
        for ext in ("jpg", "jpeg", "png"):
            image_paths.extend(self._base_dir.glob(f"**/*.{ext}"))
        image_paths_str = [str(image.relative_to(self._base_dir)) for image in image_paths]
        filepaths = pd.Series(image_paths_str)
        metadata = cast(
            pd.DataFrame,
            filepaths.str.split("/", expand=True).rename(
                columns={0: "superclass", 1: "concept", 2: "context", 3: "filename"}
            ),
        )
        metadata["filepath"] = filepaths
        metadata.sort_index(axis=1, inplace=True)
        metadata.sort_values(by=["filepath"], axis=0, inplace=True)
        metadata = self._label_encode_metadata(metadata)
        metadata.to_csv(self._metadata_path)

    @staticmethod
    def _label_encode_metadata(metadata: pd.DataFrame) -> pd.DataFrame:
        """Label encode the extracted concept/context/superclass information."""
        for col in metadata.columns:
            # Skip over filepath and filename columns
            if "file" not in col:
                # Add a new column containing the label-encoded data
                metadata[f"{col}_le"] = metadata[col].factorize()[0]
        return metadata


def _gif_to_jpeg(image: Image.Image) -> Image.Image:
    background = Image.new('RGBA', image.size, (255, 255, 255))
    return Image.alpha_composite(background, image).convert("RGB")


def preprocess_nico(path: Path) -> None:
    """
    Preprocess the original NICO data.
    This preprocessing entails two things:
        1) Renaming the files according to their concept/context/order of appearance.
        This renaming was necessary as some of the file-names were too long
        to be unzipped.
        2) Converting any GIFs into JPEGs by extracting the first frame.
    Note that this preprocessing will be performed **in-place**, overwriting the original data.
    """
    for superclass in ("animals", "vehicles"):
        superclass_dir = path / superclass
        for class_dir in superclass_dir.glob("*"):
            for context_dir in class_dir.glob("*"):
                images_paths: List[Path] = []
                for ext in ("jpg", "jpeg", "png", "gif"):
                    images_paths.extend(context_dir.glob(f"**/*.{ext}"))
                for counter, image_path in enumerate(images_paths):
                    try:
                        image = Image.open(image_path)
                        if image.format == "GIF":
                            image = image.convert("RGBA")
                            # Convert from gif to jpeg by extracting the first frame
                            new_image = _gif_to_jpeg(image)
                            new_image_path = image_path.with_suffix(".jpg")
                            # Delete the original gif
                            image_path.unlink()
                            new_image.save(new_image_path, "JPEG")
                            assert new_image_path.exists()
                            image_path = new_image_path

                        concept = image_path.parent.parent.stem
                        context = image_path.parent.stem
                        new_name = (
                            image_path.parent
                            / f"{concept}_{context}_{counter:04}{image_path.suffix}".replace(
                                " ", "_"
                            )
                        )
                        image_path.rename(new_name)
                    # Image is corrupted - delete it
                    except UnidentifiedImageError:
                        image_path.unlink()



from enum import auto
from typing import Any, Dict, List, Mapping, Protocol, TypeVar, Union, runtime_checkable
from typing_extensions import TypeAlias

import numpy as np
import numpy.typing as npt
from ranzen import StrEnum
from ranzen.torch.loss import ReductionType
from torch import Tensor
from torch.optim.lr_scheduler import CosineAnnealingWarmRestarts, ExponentialLR, StepLR
from torchmetrics import Metric

__all__ = [
    "IndexType",
    "Indexable",
    "LRScheduler",
    "Loss",
    "MetricDict",
    "NDArrayR",
    "Sized",
    "Stage",
    "METRIC_COLLECTION",
]

_NUMBER = Union[int, float]
_METRIC = Union[Metric, Tensor, _NUMBER]
METRIC_COLLECTION = Union[_METRIC, Mapping[str, _METRIC]]


class Loss(Protocol):
    reduction: Union[ReductionType, str]

    def __call__(self, input: Tensor, target: Tensor, **kwargs: Any) -> Tensor:
        ...


class Stage(StrEnum):
    FIT = auto()
    VALIDATE = auto()
    TEST = auto()


LRScheduler: TypeAlias = Union[CosineAnnealingWarmRestarts, ExponentialLR, StepLR]
MetricDict: TypeAlias = Dict[str, METRIC_COLLECTION]
NDArrayR: TypeAlias = Union[npt.NDArray[np.floating], npt.NDArray[np.integer]]
IndexType: TypeAlias = Union[int, List[int], slice]

T_co = TypeVar("T_co", covariant=True)


@runtime_checkable
class Sized(Protocol[T_co]):
    def __len__(self) -> int:
        ...


@runtime_checkable
class Indexable(Protocol[T_co]):
    def __getitem__(self, index: IndexType) -> Any:
        ...



"""Credit Dataset."""

from dataclasses import dataclass

from ethicml.data import Credit, Dataset
from ethicml.data import CreditSplits as CreditSens

from conduit.fair.data.datamodules.tabular.base import EthicMlDataModule

__all__ = ["CreditDataModule", "CreditSens"]


@dataclass(kw_only=True)
class CreditDataModule(EthicMlDataModule):
    """Data Module for the Credit Dataset."""

    sens_feat: CreditSens = CreditSens.SEX
    disc_feats_only: bool = False

    @property
    def em_dataset(self) -> Dataset:
        return Credit(
            split=self.sens_feat, discrete_only=self.disc_feats_only, invert_s=self.invert_s
        )



"""ColoredMNIST Dataset."""

from enum import auto
from pathlib import Path
from typing import ClassVar, Dict, List, Optional, Tuple, Union, cast
from typing_extensions import TypeAlias, override

from PIL import Image
import numpy as np
import numpy.typing as npt
from ranzen import StrEnum
import torch
from torch import Tensor
from torchvision.datasets import MNIST  # type: ignore

from conduit.data.structures import TernarySample
from conduit.types import IndexType, NDArrayR

from .base import CdtVisionDataset
from .utils import ImageTform, RawImage

__all__ = ["ColoredMNIST", "ColoredMNISTSplit", "MNISTColorizer"]


class MNISTColorizer:
    """Convert a greyscale MNIST image to RGB."""

    COLORS: ClassVar[Tensor] = torch.tensor(
        [
            (0, 255, 255),
            (0, 0, 255),  # blue
            (255, 0, 255),
            (0, 128, 0),
            (0, 255, 0),  # green
            (128, 0, 0),
            (0, 0, 128),
            (128, 0, 128),
            (255, 0, 0),  # red
            (255, 255, 0),  # yellow
        ],
        dtype=torch.float32,
    )

    def __init__(
        self,
        scale: float,
        *,
        min_val: float = 0.0,
        max_val: float = 1.0,
        binarize: bool = False,
        background: bool = False,
        black: bool = True,
        greyscale: bool = False,
        color_indices: Optional[Union[List[int], slice]] = None,
        seed: Optional[int] = 42,
    ) -> None:
        """
        Colorizes a grayscale image by sampling colors from multivariate normal distributions.

        The distribution is centered on predefined means and standard deviation determined by the
        scale argument.

        :param min_val: Minimum value the input data can take (needed for clamping).
        :param max_val: Maximum value the input data can take (needed for clamping).
        :param scale: Standard deviation of the multivariate normal distributions from which
            the colors are drawn. Lower values correspond to higher bias. Defaults to 0.02.
        :param binarize: Whether to the binarize the grayscale data before colorisation.
        :param background: Whether to color the background instead of the foreground.
        :param black: Whether not to invert the black. Defaults to True.
        :param greyscale: Whether to greyscale the colorised images. Defaults to False.
        :param color_indices: Choose specific colors if you don't need all 10.
        :param seed: Random seed used for sampling colors.
        """
        super().__init__()
        self.min_val = min_val
        self.max_val = max_val
        self.binarize = binarize
        self.background = background
        self.black = black
        self.greyscale = greyscale
        self.scale = scale
        self.seed = seed

        self.generator = (
            torch.default_generator
            if self.seed is None
            else torch.Generator().manual_seed(self.seed)
        )
        self.palette = self.COLORS if color_indices is None else self.COLORS[color_indices]

    def _sample_colors(self, mean_color_values: Tensor) -> Tensor:
        return (
            torch.normal(mean=mean_color_values, std=self.scale, generator=self.generator).clip(
                0, 255
            )
            / 255.0
        )

    def __call__(
        self, images: Union[Tensor, NDArrayR], *, labels: Union[Tensor, NDArrayR]
    ) -> Tensor:
        """Apply the transformation.

        :param images:  Greyscale images to be colorized. Expected to be unnormalized (in the range
            [0, 255]).
        :param labels: Indexes (0-9) indicating the gaussian distribution from which to sample each
            image's color.
        :returns: Images converted to RGB.
        """
        if isinstance(images, np.ndarray):
            images = torch.as_tensor(images, dtype=torch.float32)
        if isinstance(labels, np.ndarray):
            labels = torch.as_tensor(labels, dtype=torch.long)
        # Add a singular channel dimension if one isn't already there.
        images = cast(Tensor, torch.atleast_3d(images))
        if images.ndim == 3:
            images = images.unsqueeze(1)
        images = images.expand(-1, 3, -1, -1)

        colors = self._sample_colors(self.palette[labels]).view(-1, 3, 1, 1)

        if self.binarize:
            images = (images > 127).float()

        if self.background:
            if self.black:
                # colorful background, black digits
                images_colorized = (1 - images) * colors
            else:
                # colorful background, white digits
                images_colorized = images + colors
        elif self.black:
            # black background, colorful digits
            images_colorized = images * colors
        else:
            # white background, colorful digits
            images_colorized = 1 - images * (1 - colors)

        if self.greyscale:
            images_colorized = images_colorized.mean(dim=1, keepdim=True)

        return images_colorized


def _filter_data_by_labels(
    data: Tensor,
    *,
    targets: Tensor,
    label_map: Dict[int, int],
) -> Tuple[Tensor, Tensor]:
    final_mask = torch.zeros_like(targets).bool()
    for old_label, new_label in label_map.items():
        mask = targets == old_label
        targets[mask] = new_label
        final_mask |= mask
    return data[final_mask], targets[final_mask]


class ColoredMNISTSplit(StrEnum):
    TRAIN = auto()
    TEST = auto()


SampleType: TypeAlias = TernarySample


class ColoredMNIST(CdtVisionDataset[SampleType, Tensor, Tensor]):
    x: npt.NDArray[np.uint8]

    def __init__(
        self,
        root: Union[str, Path],
        *,
        download: bool = True,
        transform: Optional[ImageTform] = None,
        label_map: Optional[Dict[int, int]] = None,
        colors: Optional[List[int]] = None,
        num_colors: int = 10,
        scale: float = 0.2,
        correlation: Optional[float] = None,
        binarize: bool = False,
        greyscale: bool = False,
        background: bool = False,
        black: bool = True,
        split: Optional[Union[ColoredMNISTSplit, str, List[int]]] = None,
        seed: Optional[int] = 42,
    ) -> None:
        self.split = ColoredMNISTSplit(split) if isinstance(split, str) else split
        self.label_map = label_map
        self.scale = scale
        self.num_colors = num_colors
        self.colors = colors
        self.background = background
        self.binarize = binarize
        self.black = black
        self.greyscale = greyscale
        self.seed = seed
        # Note: a correlation coefficient of '1' corresponds to perfect correlation between
        # digit and class while a correlation coefficient of '-1' corresponds to perfect
        # anti-correlation.
        if correlation is None:
            correlation = 1.0 if split is ColoredMNISTSplit.TRAIN else 0.5
        if not 0 <= correlation <= 1:
            raise ValueError(
                "Strength of correlation between colour and targets must be between 0 and 1."
            )
        self.correlation = correlation

        if isinstance(self.split, ColoredMNISTSplit):
            base_dataset = MNIST(
                root=str(root), download=download, train=self.split is ColoredMNISTSplit.TRAIN
            )
            x = base_dataset.data
            y = base_dataset.targets
        else:
            x_ls, y_ls = [], []
            for _split in ColoredMNISTSplit:
                base_dataset = MNIST(
                    root=str(root), download=download, train=_split is ColoredMNISTSplit.TRAIN
                )
                x_ls.append(base_dataset.data)
                y_ls.append(base_dataset.targets)
            x = torch.cat(x_ls, dim=0)
            y = torch.cat(y_ls, dim=0)
            # custom split
            if self.split is not None:
                x = x[self.split]
                y = y[self.split]

        if self.label_map is not None:
            x, y = _filter_data_by_labels(data=x, targets=y, label_map=self.label_map)
        s = y % self.num_colors
        s_unique, s_unique_inv = s.unique(return_inverse=True)

        generator = (
            torch.default_generator
            if self.seed is None
            else torch.Generator().manual_seed(self.seed)
        )
        inv_card_s = 1 / len(s_unique)
        if self.correlation < 1:
            flip_prop = self.correlation * (1.0 - inv_card_s) + inv_card_s
            # Change the values of randomly-selected labels to values other than their original ones
            num_to_flip = round((1 - flip_prop) * len(s))
            to_flip = torch.randperm(len(s), generator=generator)[:num_to_flip]
            s_unique_inv[to_flip] += torch.randint(low=1, high=len(s_unique), size=(num_to_flip,))
            # s labels live inside the Z/(num_colors * Z) ring
            s_unique_inv[to_flip] %= len(s_unique)
            s = s_unique[s_unique_inv]

        # Convert the greyscale images of shape ( H, W ) into 'colour' images of shape ( C, H, W )
        colorizer = MNISTColorizer(
            scale=self.scale,
            background=self.background,
            black=self.black,
            binarize=self.binarize,
            greyscale=self.greyscale,
            color_indices=self.colors,
            seed=self.seed,
        )
        x_colorized = colorizer(images=x, labels=s)
        # Convert to HWC format for compatibility with transforms
        x_colorized = x_colorized.movedim(1, -1).numpy().astype(np.uint8)

        super().__init__(x=x_colorized, y=y, s=s, transform=transform, image_dir=root)

    @override
    def _load_image(self, index: IndexType) -> RawImage:
        image = self.x[index]
        if self._il_backend == "pillow":
            return Image.fromarray(image)
        return image



"""Base class from which all data-modules in conduit inherit."""

from abc import abstractmethod
from dataclasses import dataclass, field
import logging
from typing import Any, Generic, List, Optional, Sequence, Tuple, TypeVar, Union, cast, final
from typing_extensions import Self, override

import pytorch_lightning as pl
from ranzen.torch import SequentialBatchSampler, StratifiedBatchSampler, TrainingMode
from ranzen.torch.data import num_batches_per_epoch
import torch
from torch.utils.data import Sampler

from conduit.data.datasets.base import CdtDataset
from conduit.data.datasets.utils import CdtDataLoader, extract_base_dataset, get_group_ids
from conduit.data.datasets.wrappers import InstanceWeightedDataset
from conduit.data.structures import Dataset, SampleBase, SizedDataset, TrainValTestSplit
from conduit.logging import init_logger
from conduit.types import Stage

__all__ = ["CdtDataModule"]

D = TypeVar("D", bound=SizedDataset)
I = TypeVar("I", bound=SampleBase, covariant=True)


@dataclass(kw_only=True, eq=False)
class CdtDataModule(pl.LightningDataModule, Generic[D, I]):
    """Base DataModule for both Tabular and Vision data-modules.

    :param val_prop: Proportion of samples to designate as the validation split.
    :param test_prop: Proportion of samples to designate as the test split.

    :param num_workers: how many subprocesses to use for data-loading.
        ``0`` means that the data will be loaded in the main process.

    :param train_batch_size: Batch size to use for the training data.
    :param eval_batch_size: Batch size to use for the validationa and test data.
    :param seed: PRNG seed to use for splitting the data.
    :param persist_workers: Whether to persistent workers in dataloader.

    :param pin_memory: If ``True``, the data loader will copy Tensors
        into CUDA pinned memory before returning them.  If your data elements
        are a custom type, or your :attr:`collate_fn` returns a batch that is a custom type,
        see the example below.

    :param stratified_sampling: Whether to draw class-balanced batches of data with the
        train-dataloader (only applicable to classification datasets with discrete ``y`` values).

    :param instance_weighting: Whether to instance-weight samples of the dataset (only applicable to
        classification datasets with discrete ``y`` values).

    :param training_mode: Which mode to use for sampling with the train-dataloader. If `epoch` then
        the train-dataloader will be exhausted after a complete pass though the data. If `step` then
        samples will be drawn ad infinitum by the dataloader.
    """

    train_batch_size: int = 64
    eval_batch_size: Optional[int] = field(default=None)
    val_prop: float = 0.2
    test_prop: float = 0.2
    num_workers: int = 0
    seed: int = 47
    persist_workers: bool = False
    pin_memory: bool = True
    stratified_sampling: bool = False
    instance_weighting: bool = False
    training_mode: TrainingMode = TrainingMode.epoch

    _logger: Optional[logging.Logger] = field(default=None, init=False)

    _train_data_base: Optional[Dataset] = field(default=None, init=False)
    _val_data_base: Optional[Dataset] = field(default=None, init=False)
    _test_data_base: Optional[Dataset] = field(default=None, init=False)

    _train_data: Union[None, D, InstanceWeightedDataset] = field(default=None, init=False)
    _val_data: Optional[D] = field(default=None, init=False)
    _test_data: Optional[D] = field(default=None, init=False)
    _card_s: Optional[int] = field(default=None, init=False)
    _card_y: Optional[int] = field(default=None, init=False)
    _dim_s: Optional[torch.Size] = field(default=None, init=False)
    _dim_y: Optional[torch.Size] = field(default=None, init=False)
    _dim_x: Optional[Tuple[int, ...]] = field(default=None, init=False)

    @final
    def __new__(cls, *args: Any, **kwargs: Any) -> Self:
        obj = object.__new__(cls)
        pl.LightningDataModule.__init__(obj)
        return obj

    @property
    def eval_bs(self) -> int:
        if self.eval_batch_size is None:
            return self.train_batch_size
        return self.eval_batch_size

    @property
    def logger(self) -> logging.Logger:
        if self._logger is None:
            self._logger = init_logger(self.__class__.__name__)
        return self._logger

    @property
    def train_prop(self) -> float:
        return 1 - (self.val_prop + self.test_prop)

    def make_dataloader(
        self,
        ds: D,
        *,
        batch_size: int,
        shuffle: bool = False,
        drop_last: bool = False,
        batch_sampler: Optional[Sampler[List[int]]] = None,
    ) -> CdtDataLoader[I]:
        """Make DataLoader."""
        return CdtDataLoader(
            ds,
            batch_size=batch_size if batch_sampler is None else 1,
            shuffle=shuffle,
            num_workers=self.num_workers,
            pin_memory=self.pin_memory,
            drop_last=drop_last,
            persistent_workers=self.persist_workers,
            batch_sampler=batch_sampler,
        )

    @property
    @final
    def train_data_base(self) -> D:
        self._check_setup_called("train_data_base")
        return cast(D, self._train_data_base)

    @property
    @final
    def train_data(self) -> D:
        self._check_setup_called()
        return cast(D, self._train_data)

    @property
    @final
    def val_data(self) -> D:
        self._check_setup_called()
        return cast(D, self._val_data)

    @property
    @final
    def test_data(self) -> D:
        self._check_setup_called()
        return cast(D, self._test_data)

    def train_dataloader(
        self, *, shuffle: bool = False, drop_last: bool = False, batch_size: Optional[int] = None
    ) -> CdtDataLoader[I]:
        batch_size = self.train_batch_size if batch_size is None else batch_size

        if self.stratified_sampling:
            group_ids = get_group_ids(self.train_data)
            num_groups = len(group_ids.unique())
            num_samples_per_group = batch_size // num_groups
            if batch_size % num_groups:
                self.logger.info(
                    "For stratified sampling, the batch size must be a multiple of the number of"
                    " groups.Since the batch size is not integer divisible by the number of groups"
                    f" ({num_groups}),the batch size is being reduced to"
                    f" {num_samples_per_group * num_groups}."
                )
            batch_sampler = StratifiedBatchSampler(
                group_ids=group_ids.squeeze().tolist(),
                num_samples_per_group=num_samples_per_group,
                shuffle=shuffle,
                base_sampler="sequential",
                training_mode=self.training_mode,
                drop_last=drop_last,
            )
        else:
            batch_sampler = SequentialBatchSampler(
                data_source=self._train_data,  # type: ignore
                batch_size=batch_size,
                shuffle=shuffle,
                training_mode=self.training_mode,
                drop_last=drop_last,
            )
        return self.make_dataloader(
            ds=self.train_data, batch_size=self.train_batch_size, batch_sampler=batch_sampler
        )

    @override
    def val_dataloader(self) -> CdtDataLoader[I]:
        return self.make_dataloader(batch_size=self.eval_bs, ds=self.val_data)

    @override
    def test_dataloader(self) -> CdtDataLoader[I]:
        return self.make_dataloader(batch_size=self.eval_bs, ds=self.test_data)

    @property
    def dim_x(self) -> Sequence[int]:
        """
        Returns the dimensions of the first input (x).

        :returns: Tuple containing the dimensions of the (first) input.
        """
        if self._dim_x is None:
            self._check_setup_called()
            input_size = tuple(self._train_data[0].x.shape)  # type: ignore
            self._dim_x = input_size
        return self._dim_x

    def size(self) -> Sequence[int]:
        """Alias for ``dim_x``.

        :returns: Tuple containing the dimensions of the (first) input.
        """
        return self.dim_x

    @final
    def _num_samples(self, dataset: D) -> int:
        if hasattr(dataset, "__len__"):
            return len(dataset)  # type: ignore
        raise AttributeError(
            "Number of samples cannot be determined as dataset of type"
            f" '{dataset.__class__.__name__}' has no '__len__' attribute defined."
        )

    @property
    @final
    def num_train_samples(self) -> int:
        return self._num_samples(self.train_data)

    @property
    @final
    def num_val_samples(self) -> int:
        return self._num_samples(self.val_data)

    @property
    @final
    def num_test_samples(self) -> int:
        return self._num_samples(self.test_data)

    @final
    def num_train_batches(self, drop_last: bool = False) -> int:
        if self.training_mode is TrainingMode.step:
            raise AttributeError(
                "'num_train_batches' can only be computed when 'training_mode' is set to 'epoch'."
            )
        return num_batches_per_epoch(
            num_samples=self.num_train_samples,
            batch_size=self.train_batch_size,
            drop_last=drop_last,
        )

    @property
    @final
    def dim_y(self) -> Tuple[int, ...]:
        self._check_setup_called()
        return self._get_base_dataset().dim_y

    @property
    @final
    def dim_s(self) -> Tuple[int, ...]:
        self._check_setup_called()
        return self._get_base_dataset().dim_s

    @property
    @final
    def card_y(self) -> int:
        self._check_setup_called()
        return self._get_base_dataset().card_y

    @property
    @final
    def card_s(self) -> int:
        self._check_setup_called()
        return self._get_base_dataset().card_s

    def _check_setup_called(self, caller: Optional[str] = None) -> None:
        if not self.is_set_up:
            if caller is None:
                # inspect the call stack to find out who called this function
                import inspect

                caller = inspect.getouterframes(inspect.currentframe(), 2)[1][3]

            cls_name = self.__class__.__name__
            raise AttributeError(
                f"'{cls_name}.{caller}' cannot be accessed as '{cls_name}.setup()' has "
                "not yet been called."
            )

    def _get_base_dataset(self) -> CdtDataset:
        if not isinstance(self._train_data_base, CdtDataset):
            # inspect the call stack to find out who called this function
            import inspect

            caller = inspect.getouterframes(inspect.currentframe(), 2)[1][3]
            raise AttributeError(
                f"'{caller}' can only be determined for {CdtDataset.__name__} instances."
            )
        return self._train_data_base

    @abstractmethod
    def _get_splits(self) -> TrainValTestSplit[D]:
        ...

    @property
    def is_set_up(self) -> bool:
        return self._train_data is not None

    def _setup(self, stage: Optional[Stage] = None) -> None:
        train_data: Union[None, D, InstanceWeightedDataset]
        train_data, self._val_data, self._test_data = self._get_splits()
        if self.instance_weighting:
            train_data = InstanceWeightedDataset(train_data)
        self._train_data = train_data

    def _post_setup(self) -> None:
        # Make information (cardinality/dimensionality) about the dataset directly accessible through the data-module
        self._train_data_base = extract_base_dataset(
            dataset=self.train_data, return_subset_indices=False
        )
        self._val_data_base = extract_base_dataset(
            dataset=self.val_data, return_subset_indices=False
        )
        self._test_data_base = extract_base_dataset(
            dataset=self.test_data, return_subset_indices=False
        )

    @override
    @final
    def setup(  # type: ignore
        self,
        stage: Optional[Stage] = None,
        force_reset: bool = False,
    ) -> None:
        # Only perform the setup if it hasn't already been done
        if force_reset or (not self.is_set_up):
            self._setup(stage=stage)
            self._post_setup()



from enum import Enum
from typing import Callable, Optional, Protocol, TypeVar, Union, cast
from typing_extensions import Self

import torch
from torch import Tensor
from torch.autograd.function import Function, NestedIOFunction
import torch.distributed
import torch.nn as nn
import torch.nn.functional as F

__all__ = [
    "DecoupledContrastiveLoss",
    "SupConReduction",
    "decoupled_contrastive_loss",
    "moco_v2_loss",
    "simclr_loss",
    "soft_supcon_loss",
    "supcon_loss",
]


class _BatchStore(Protocol):
    batch_size: int


class _Synchronize(Function):
    @staticmethod
    def forward(ctx: NestedIOFunction, tensor: Tensor) -> Tensor:  # type: ignore
        context = cast(_BatchStore, ctx)
        context.batch_size = tensor.shape[0]

        gathered_tensor = [
            torch.zeros_like(tensor) for _ in range(torch.distributed.get_world_size())
        ]

        torch.distributed.all_gather(gathered_tensor, tensor)
        gathered_tensor = torch.cat(gathered_tensor, dim=0)

        return gathered_tensor

    @staticmethod
    def backward(ctx: NestedIOFunction, grad_output: Tensor) -> Tensor:  # type: ignore
        context = cast(_BatchStore, ctx)
        grad_input = grad_output.clone()
        torch.distributed.all_reduce(grad_input, op=torch.distributed.ReduceOp.SUM, async_op=False)
        idx_from = torch.distributed.get_rank() * context.batch_size
        idx_to = (torch.distributed.get_rank() + 1) * context.batch_size
        return grad_input[idx_from:idx_to]


def maybe_synchronize(input: Tensor) -> Tensor:
    if torch.distributed.is_available() and torch.distributed.is_initialized():
        return _Synchronize.apply(input)  # type: ignore
    return input


def logsumexp(
    input: Tensor, *, dim: int, keepdim: bool = False, keep_mask: Optional[Tensor] = None
) -> Tensor:
    """Numerically stable implementation of logsumexp that allows for masked summation."""
    if keep_mask is None:
        return input.logsumexp(dim=dim, keepdim=keepdim)
    eps = torch.finfo(input.dtype).eps
    max_offset = eps * keep_mask.to(input.dtype)
    max_ = torch.max(input + max_offset, dim=dim, keepdim=True).values
    input = input - max_
    if keepdim is False:
        max_ = max_.squeeze(dim)
    input_exp_m = input.exp() * keep_mask
    return max_ + torch.log(input_exp_m.sum(dim=dim, keepdim=keepdim) + eps)


def moco_v2_loss(
    anchors: Tensor,
    *,
    positives: Tensor,
    negatives: Tensor,
    temperature: Union[float, Tensor] = 1.0,
    dcl: bool = True,
    normalize: bool = True,
) -> Tensor:
    """
    Loss function used in the training of MoCo v2.

    :param anchors: Logits of the anchor samples.
    :param positives: Logits of the positive samples w.r.t. ``anchors`` (derived from different
        views of the same samples).
    :param negatives: Logits of the negative samples w.r.t ``anchors`` (derived from different
        samples).
    :param temperature: Inverse temperature parameter used for adjusting the sharpness of the
        softmax distribution. Smaller values lead to hard negatives being overemphasized.

    :param dcl: Whether to decouple the numerators from the denominators, meaning that the
        cross-view distances (similarity between the anchors and positives) are not included in the
        negative component of the loss. This has been shown to improve convergence and stability,
        particularly when working with smaller batch sizes.

    :param normalize: Whether to l2 normalize the logits before computing the inter-sample
        similarities.

    :returns: Loss as a 0-dimensional tensor.
    """
    if normalize:
        anchors = F.normalize(anchors, dim=1, p=2)
        positives = F.normalize(positives, dim=1, p=2)
        negatives = F.normalize(negatives, dim=1, p=2)

    positives = maybe_synchronize(positives)
    negatives = maybe_synchronize(negatives)
    if positives.requires_grad or negatives.requires_grad:
        anchors = maybe_synchronize(anchors)

    n, d = anchors.size(0), anchors.size(-1)
    anchors = anchors.view(n, -1, d)
    positives = positives.view(n, 1, d)
    l_pos = (anchors * positives).sum(-1).view(-1, 1) / temperature
    l_neg = (anchors @ negatives.T).view(l_pos.size(0), -1) / temperature
    # Compute the partition function either according to the original InfoNCE formulation
    # or according to the DCL formulation which excludes the positive samples.
    z = l_neg.logsumexp(dim=1) if dcl else torch.cat([l_pos, l_neg], dim=1).logsumexp(dim=1)
    return (z - l_pos).mean()


def simclr_loss(
    anchors: Tensor,
    *,
    targets: Tensor,
    temperature: Union[float, Tensor] = 1.0,
    dcl: bool = True,
    normalize: bool = True,
) -> Tensor:
    if normalize:
        anchors = F.normalize(anchors, dim=1, p=2)
        targets = F.normalize(targets, dim=1, p=2)

    if anchors.requires_grad:
        targets = maybe_synchronize(targets)
    if targets.requires_grad:
        anchors = maybe_synchronize(anchors)

    logits = (anchors @ targets.T) / temperature
    pos_idxs = torch.arange(logits.size(0), device=logits.device).view(
        -1, *((1,) * (logits.ndim - 1))
    )
    l_pos = logits.gather(-1, pos_idxs)

    z_mask = None
    if dcl:
        z_mask = ~torch.eye(len(logits), dtype=torch.bool, device=logits.device)
        if anchors.ndim == 3:
            z_mask = z_mask.unsqueeze(1)
    z = logsumexp(logits, dim=-1, keep_mask=z_mask)
    return (z.sum() - l_pos.sum()) / z.numel()


T = TypeVar("T", Tensor, None)


class SupConReduction(Enum):
    MEAN = "mean"
    SUM = "sum"


def supcon_loss(
    anchors: Tensor,
    *,
    anchor_labels: Tensor,
    candidates: T = None,
    candidate_labels: T = None,
    temperature: Union[float, Tensor] = 0.1,
    exclude_diagonal: bool = False,
    dcl: bool = True,
    margin: float = 0,
    reduction: SupConReduction = SupConReduction.MEAN,
    quantile_threshold: float = 0.0,
    normalize: bool = True,
) -> Tensor:
    if margin < 0:
        raise ValueError("'margin' must be non-negative.")
    if len(anchors) != len(anchor_labels):
        raise ValueError("'anchors' and 'anchor_labels' must match in size at dimension 0.")
    if not (0 <= quantile_threshold < 1):
        raise ValueError("'quantile_threshold' must be in the range [0, 1).")

    # Create new variables for the candidate- variables to placate
    # the static-type checker.
    if normalize:
        anchors = F.normalize(anchors, dim=1, p=2)
    if candidates is None:
        candidates_t = anchors
        candidate_labels_t = anchor_labels
        # Forbid interactions between the samples and themsleves.
        exclude_diagonal = True
    else:
        candidates_t = candidates
        candidate_labels_t = cast(Tensor, candidate_labels)
        if len(candidates_t) != len(candidate_labels_t):
            raise ValueError(
                "'candidates' and 'candidate_labels' must match in size at dimension 0."
            )
        if normalize:
            candidates_t = F.normalize(candidates_t, dim=1, p=2)
        candidates_t = maybe_synchronize(candidates_t)
        candidate_labels_t = maybe_synchronize(candidate_labels_t)
        # If the gradient is to be computed bi-directionally then both the queries and the keys
        # need to be collated across all devices (otherwise just doing so for the keys is
        # sufficient).
        if candidates_t.requires_grad:
            anchors = maybe_synchronize(anchors)
            anchor_labels = maybe_synchronize(anchor_labels)

    anchor_labels = anchor_labels.view(-1, 1)
    candidate_labels_t = candidate_labels_t.flatten()
    # The positive samples for a given anchor are those samples from the candidate set sharing its
    # label.
    pos_mask = anchor_labels == candidate_labels_t
    neg_mask = None
    if dcl:
        neg_mask = ~pos_mask
    elif exclude_diagonal:
        neg_mask = torch.full_like(pos_mask, True).fill_diagonal_(False)
    if exclude_diagonal:
        pos_mask.fill_diagonal_(False)

    pos_inds = pos_mask.nonzero(as_tuple=True)
    row_inds, col_inds = pos_inds
    # Return early if there are no positive pairs.
    if len(row_inds) == 0:
        return anchors.new_zeros(())
    # Only compute the pairwise similarity for those samples which have positive pairs.
    selected_rows, row_inverse, row_counts = row_inds.unique(
        return_inverse=True, return_counts=True
    )
    logits = anchors[selected_rows] @ candidates_t.T / temperature
    # Subtract the maximum for numerical stability.
    logits_max = logits.max(dim=1, keepdim=True).values
    logits -= logits_max

    if margin > 0:
        logits[row_inverse, ..., col_inds] -= margin
    # Tile the row counts if dealing with multicropping.
    positives = logits[row_inverse, ..., col_inds].flatten()
    if reduction is SupConReduction.MEAN:
        if anchors.ndim == 3:
            row_counts = row_counts.unsqueeze(1).expand(-1, anchors.size(1))
        counts_flat = row_counts[row_inverse].flatten()
        positives /= counts_flat

    if neg_mask is not None:
        neg_mask = neg_mask[selected_rows]
        if anchors.ndim == 3:
            neg_mask = neg_mask.unsqueeze(1)
    if quantile_threshold > 0:
        quantiles = torch.quantile(logits.float(), q=quantile_threshold, dim=-1, keepdim=True)
        neg_mask &= logits >= quantiles
    z = logsumexp(logits, dim=-1, keep_mask=neg_mask)
    if reduction is SupConReduction.SUM:
        z *= row_counts.view(*((-1,) + ((z.ndim - 1) * (1,))))
    return (z.sum() - positives.sum()) / z.numel()


def soft_supcon_loss(
    z1: Tensor,
    *,
    p1: Tensor,
    z2: T = None,
    p2: T = None,
    temperature: Union[float, Tensor] = 0.1,
    exclude_diagonal: bool = False,
    dcl: bool = True,
    normalize: bool = True,
) -> Tensor:
    if len(z1) != len(p1):
        raise ValueError("'z1' and 'p1' must match in size at dimension 0.")
    # Create new variables for the candidate- variables to placate
    # the static-type checker.
    if normalize:
        z1 = F.normalize(z1, dim=1, p=2)
    if z2 is None:
        z2_t = z1
        p2_t = p1
        # Forbid interactions between the samples and themsleves.
        exclude_diagonal = True
    else:
        z2_t = z2
        p2_t = cast(Tensor, p2)
        if len(z2_t) != len(p2_t):
            raise ValueError("'z2' and 'p2' must match in size at dimension 0.")
        if p1.size(1) != p2_t.size(1):
            raise ValueError("'p1' and 'p2' must match in size at dimension 1.")
        if normalize:
            z2_t = F.normalize(z2_t, dim=1, p=2)
        z2_t = maybe_synchronize(z2_t)
        p2_t = maybe_synchronize(p2_t)
        if z2_t.requires_grad:
            z1 = maybe_synchronize(z1)
            p1 = maybe_synchronize(p1)

    y1 = torch.ceil(p1).long()
    y2 = torch.ceil(p2_t).long()
    # The positive samples for a given anchor are those samples from the candidate set sharing its
    # label.
    mask = y1.unsqueeze(1) == y2
    diag = None
    if exclude_diagonal:
        diag = mask.new_zeros((len(z1), len(z2_t)), dtype=torch.bool).fill_diagonal_(True)
        mask = mask ^ diag.unsqueeze(-1)

    pos_inds = mask.nonzero(as_tuple=True)
    row_inds, col_inds, coupling_inds = pos_inds
    # Return early if there are no positive pairs.
    if len(row_inds) == 0:
        return z1.new_zeros(())
    # Only compute the pairwise similarity for those samples which have positive pairs.
    selected_rows, row_inverse, _ = row_inds.unique(return_inverse=True, return_counts=True)
    logits = z1[selected_rows] @ z2_t.T
    logits = logits / temperature
    coupling_coeffs = p1[row_inverse, ..., coupling_inds] * p2_t[col_inds, ..., coupling_inds]
    denom = coupling_coeffs.new_zeros(len(selected_rows)).scatter_(
        dim=0, index=row_inverse, src=coupling_coeffs, reduce="add"
    )[row_inverse]
    positives = logits[row_inverse, ..., col_inds]
    weighted_positives = (positives * coupling_coeffs) / denom

    neg_mask = None
    if diag is not None:
        neg_mask = ~diag[selected_rows]

    z = logsumexp(logits, dim=-1, keep_mask=neg_mask)
    return (z.sum() - weighted_positives.sum()) / z.numel()


def decoupled_contrastive_loss(
    z1: Tensor,
    z2: Tensor,
    *,
    temperature: float = 0.1,
    weight_fn: Optional[Callable[[Tensor, Tensor], Tensor]] = None,
) -> Tensor:
    """
    Calculates the one-way `decoupled constrastive loss <https://arxiv.org/pdf/2110.06848.pdf>`_.

    :param z1: First embedding vector
    :param z2: Second embedding vector
    :param weight_fn: The weighting function of the positive sample loss.
    :param temperature: Temperature controlling the sharpness of the softmax distribution.

    :return: One-way loss between the embedding vectors.
    """
    cross_view_distance = torch.mm(z1, z2.t())
    positive_loss = -torch.diag(cross_view_distance) / temperature
    if weight_fn is not None:
        positive_loss = positive_loss * weight_fn(z1, z2)
    neg_similarity = torch.cat((z1 @ z1.t(), cross_view_distance), dim=1) / temperature
    neg_mask = torch.eye(z1.size(0), device=z1.device).repeat(1, 2)
    eps = torch.finfo(z1.dtype).eps
    negative_loss = torch.logsumexp(neg_similarity + neg_mask * eps, dim=1, keepdim=False)
    return (positive_loss + negative_loss).mean()


def _von_mises_fisher_weighting(z1: Tensor, z2: Tensor, *, sigma: float = 0.5) -> Tensor:
    return 2 - len(z1) * ((z1 * z2).sum(dim=1) / sigma).softmax(dim=0).squeeze()


class DecoupledContrastiveLoss(nn.Module):
    """
    Implementation of the Decoupled Contrastive Loss proposed in
    https://arxiv.org/pdf/2110.06848.pdf, adapted from the `official code
    <https://github.com/raminnakhli/Decoupled-Contrastive-Learning>`_
    """

    def __init__(
        self,
        temperature: float = 0.1,
        *,
        weight_fn: Optional[Callable[[Tensor, Tensor], Tensor]] = None,
    ) -> None:
        """
        :param weight_fn: The weighting function of the positive sample loss.
        :param temperature: Temperature controlling the sharpness of the softmax distribution.
        """
        super().__init__()
        self.temperature = temperature
        self.weight_fn = weight_fn

    def forward(self, z1: Tensor, z2: Tensor) -> Tensor:
        """
        Calculates the one-way decoupled constrastive loss.

        :param z1: First embedding vector
        :param z2: Second embedding vector
        :return: One-way loss between the embedding vectors.
        """
        return decoupled_contrastive_loss(
            z1, z2, temperature=self.temperature, weight_fn=self.weight_fn
        )

    @classmethod
    def with_vmf_weighting(cls, sigma: float = 0.5, *, temperature: float = 0.1) -> Self:
        """
        Initialise the DCL loss with von Mises-Fisher weighting.

        :param sigma: :math:`\\sigma` (scale) parameter for the weigting function.
        :param temperature: Temperature controlling the sharpness of the softmax distribution.
        :returns: the loss object
        """
        return cls(temperature=temperature, weight_fn=_von_mises_fisher_weighting)



from functools import reduce
import operator
from pathlib import Path
from typing import List, Literal, Optional, Sequence, Union, cast, overload
from typing_extensions import Self, TypeAlias, override

import numpy as np
import numpy.typing as npt
from ranzen.types import Addable
import torch
from torch import Tensor

from conduit.data.datasets.base import CdtDataset, I, S, Y
from conduit.data.datasets.vision.utils import (
    ImageLoadingBackend,
    ImageTform,
    RawImage,
    apply_image_transform,
    img_to_tensor,
    infer_il_backend,
    load_image,
)
from conduit.data.structures import TargetData
from conduit.types import IndexType

__all__ = ["CdtVisionDataset"]

ItemType: TypeAlias = Union[RawImage, Tensor]


class CdtVisionDataset(CdtDataset[I, npt.NDArray[np.bytes_], Y, S]):
    def __init__(
        self,
        *,
        x: npt.NDArray[np.bytes_],
        image_dir: Union[Path, str],
        y: Optional[TargetData] = None,
        s: Optional[TargetData] = None,
        transform: Optional[ImageTform] = None,
    ) -> None:
        super().__init__(x=x, y=y, s=s)
        if isinstance(image_dir, str):
            image_dir = Path(image_dir)
        self.image_dir = image_dir
        self.transform = transform
        # infer the appropriate image-loading backend based on the type of 'transform'
        self._il_backend: ImageLoadingBackend = infer_il_backend(transform)

    def update_il_backend(self, transform: Optional[ImageTform]) -> None:
        new_backend = infer_il_backend(transform)
        if new_backend != self._il_backend:
            self._il_backend = new_backend
            self.logger.info(f"Using {self._il_backend} as backend for image-loading.")

    def __repr__(self) -> str:
        head = "Dataset " + self.__class__.__name__
        body = [
            f"Number of datapoints: {len(self)}",
            f"Base image-directory location: {self.image_dir.resolve()}",
            *self.extra_repr().splitlines(),
        ]
        if hasattr(self, "transform") and self.transform is not None:
            body += [repr(self.transform)]
        lines = [head] + [" " * self._repr_indent + line for line in body]
        return '\n'.join(lines)

    def _load_image(self, index: IndexType) -> RawImage:
        filepath = cast(str, self.x[index])
        return load_image(self.image_dir / filepath, backend=self._il_backend)

    @overload
    def _sample_x(self, index: int, *, coerce_to_tensor: Literal[True]) -> Tensor:
        ...

    @overload
    def _sample_x(
        self, index: Union[List[int], slice], *, coerce_to_tensor: Literal[True]
    ) -> Union[Tensor, Sequence[Tensor]]:
        ...

    @overload
    def _sample_x(self, index: int, *, coerce_to_tensor: Literal[False] = ...) -> ItemType:
        ...

    @overload
    def _sample_x(
        self, index: Union[List[int], slice], *, coerce_to_tensor: Literal[False] = ...
    ) -> Union[ItemType, Sequence[ItemType]]:
        ...

    @override
    def _sample_x(
        self, index: IndexType, *, coerce_to_tensor: bool = False
    ) -> Union[ItemType, Sequence[ItemType]]:
        if isinstance(index, slice):
            index = list(range(len(self)))[index]
        if isinstance(index, list):
            sample_ls: List[ItemType] = [
                self._sample_x(index=i, coerce_to_tensor=coerce_to_tensor)  # type: ignore
                for i in index
            ]
            elem = sample_ls[0]
            if isinstance(elem, Addable):
                summed = reduce(operator.add, sample_ls)
                return cast(ItemType, summed)
            if isinstance(elem, Tensor):
                return torch.stack(cast(List[Tensor], sample_ls), dim=0)
            elif isinstance(sample_ls[0], np.ndarray):
                return np.stack(cast(List[npt.NDArray[np.integer]], sample_ls), axis=0)
            return sample_ls

        image = self._load_image(index)
        image = apply_image_transform(image=image, transform=self.transform)
        if coerce_to_tensor and (not isinstance(image, Tensor)):
            if isinstance(image, Sequence):
                image = [img_to_tensor(subimage) for subimage in image]
            else:
                image = img_to_tensor(image)
        return image

    @override
    def subset(
        self,
        indices: Union[List[int], npt.NDArray[np.uint64], Tensor, slice],
        *,
        deep: bool = False,
        transform: Optional[ImageTform] = None,
    ) -> Self:
        """Create a subset of the dataset from the given indices.

        :param indices: The sample-indices from which to create the subset. In the case of being a
            numpy array or tensor, said array or tensor must be 0- or 1-dimensional.

        :param deep: Whether to create a copy of the underlying dataset as a basis for the subset.
            If False then the data of the subset will be a view of original dataset's data.

        :param transform: Image transform to assign to the resulting subset.

        :returns: A subset of the dataset from the given indices.
        """
        subset = super().subset(indices=indices, deep=deep)
        if transform is not None:
            subset.transform = transform
        return subset



import inspect
from typing import Any, Callable, Dict, List, TypeVar

from pytorch_lightning.utilities.types import STEP_OUTPUT
from ranzen.misc import gcopy
import torch
from torch import Tensor, nn

__all__ = [
    "PartialModule",
    "aggregate_over_epoch",
    "decorate_all_methods",
    "make_no_grad",
    "prefix_keys",
]
EPOCH_OUTPUT = List[STEP_OUTPUT]


@torch.no_grad()  # pyright: ignore
def aggregate_over_epoch(outputs: EPOCH_OUTPUT, *, metric: str) -> Tensor:
    return torch.cat([step_output[metric] for step_output in outputs])  # type: ignore


def prefix_keys(dict_: Dict[str, Any], *, prefix: str, sep: str = "/") -> Dict[str, Any]:
    """Prepend the prefix to all keys of the dict"""
    return {f"{prefix}{sep}{key}": value for key, value in dict_.items()}


T = TypeVar("T")


def decorate_all_methods(decorator: Callable[[Callable], Callable], obj: T) -> T:
    methods = inspect.getmembers(obj, inspect.ismethod)
    for name, method in methods:
        setattr(obj, name, decorator(method))
    return obj


U = TypeVar("U", bound=nn.Module)


def make_no_grad(module: U) -> U:
    return decorate_all_methods(torch.no_grad(), gcopy(module, deep=False))


class PartialModule(nn.Module):
    def __init__(self, func: Callable, **kwargs: Any):
        super().__init__()
        self.func = func
        self.bound_kwargs = kwargs

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        return self.func(*args, **kwargs, **self.bound_kwargs)



from typing import Dict, Union
from typing_extensions import TypeAlias

from torch.optim.lr_scheduler import CosineAnnealingWarmRestarts, ExponentialLR, StepLR

from .types import METRIC_COLLECTION

__all__ = [
    "LRScheduler",
    "MetricDict",
]


LRScheduler: TypeAlias = Union[CosineAnnealingWarmRestarts, ExponentialLR, StepLR]
MetricDict: TypeAlias = Dict[str, METRIC_COLLECTION]
